# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.



import unittest
from unittest.mock import patch
import pymonetdb


class MultilineResponseTest(unittest.TestCase):
    """MonetDB sometimes sends back multi-line responses. Most notably when there
       are more than one concurrent update transactions due to the Optimistic
       Concurrency Control that MonetDB is implementing, only one of them will
       succeed. In auto-commit mode the failed transactions will get something
       like the following response:

       &2 1 -1\n!4000!COMMIT: transaction is aborted because of concurrency conflicts, will ROLLBACK instead\n

       The first line is emitted by the MonetDB server when the update is
       actually executed inside the MAL plan while the second is emitted
       outside the MAL plan during the commit that fails. We should take this
       into account.

       Another case is comments that start with the '#' char and should be
       ignored. These are useful for debugging the server.

       This class is intended to create a test suite for this kind of cases.
       Since there is no deterministic way of forcing transaction failure, the
       MonetDB server will be mocked using https://pypi.python.org/pypi/mock

    """

    @patch('pymonetdb.mapi.Connection._putblock')
    @patch('pymonetdb.mapi.Connection._getblock_raw')
    def test_failed_transactions(self, mock_getblock_raw, _):
        """This test mocks two low level methods in the mapi.Connection class:
           mapi.Connection._getblock_raw
           mapi.Connection._putblock

           and tests mapi.Connection.cmd. Specifically we test for the event
           that a transaction has failed due to concurrency conflicts.
        """
        query_text = 'sINSERT INTO tbl VALUES (1)'
        response = b"&2 1 -1\n!40000!COMMIT: transaction is aborted " \
                   b"because of concurrency conflicts, will ROLLBACK instead\n"

        def mocked_getblock_raw(buf, off):
            buf[off:] = response
            return off + len(response)

        mock_getblock_raw.side_effect = mocked_getblock_raw
        c = pymonetdb.mapi.Connection()

        # Simulate a connection
        c.state = pymonetdb.mapi.STATE_READY

        # Make sure that cmd raises the correct exception
        self.assertRaises(pymonetdb.IntegrityError, c.cmd, [query_text])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import unittest
import pymonetdb
from tests.util import test_args


class TestExceptions(unittest.TestCase):
    def setUp(self):
        self.con = pymonetdb.connect(**test_args)
        cursor = self.con.cursor()
        cursor.execute('create table exceptions (s VARCHAR(1000) UNIQUE)')

    def tearDown(self):
        self.con.rollback()

    def test_unique_contraint_violated(self):
        """
        M0M29!INSERT INTO: UNIQUE constraint violated
        """
        cursor = self.con.cursor()
        x = u"something"
        cursor.execute(u'insert into exceptions VALUES (%s)', (x,))
        with self.assertRaises(pymonetdb.exceptions.IntegrityError):
            cursor.execute(u'insert into exceptions VALUES (%s)', (x,))

    def test_unique_contraint_violated2(self):
        """
        '42S02!' no such table
        """
        cursor = self.con.cursor()

        self.assertRaises(pymonetdb.exceptions.OperationalError,
                          cursor.execute,
                          u'select id from thistableshouldnotexist limit 1')



"""
This is the python implementation of the mapi protocol.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import codecs
from io import BufferedIOBase, StringIO
import os
from pathlib import Path
from shutil import copyfileobj, copyfile
import signal
import sys
from tempfile import mkdtemp
from threading import Condition, Thread
import time
from typing import Optional, Tuple
from unittest import TestCase, skipUnless


from pymonetdb import connect, Error as MonetError
from pymonetdb.exceptions import OperationalError, ProgrammingError
from pymonetdb import Download, Downloader, Upload, Uploader
from pymonetdb.filetransfer.directoryhandler import SafeDirectoryHandler, lookup_compression_algorithm
from pymonetdb.filetransfer.uploads import NormalizeCrLf
from tests.util import have_lz4, test_args, test_full


class MyException(Exception):
    pass


class MyUploader(Uploader):
    rows: int = 5_000
    error_at: Optional[int] = None
    chunk_size: int = 10_000
    force_binary: bool = False
    forget_to_return_after_error: bool = False
    do_nothing_at_all: bool = False
    ignore_cancel: bool = False
    #
    cancelled_at: Optional[int] = None

    def handle_upload(self, upload: Upload, filename: str, text_mode: bool, skip_amount: int):  # noqa: C901
        if self.do_nothing_at_all:
            return
        elif filename.startswith("x"):
            upload.send_error("filename must not start with 'x'")
            if not self.forget_to_return_after_error:
                return

        iter = range(skip_amount + 1, self.rows + 1)
        upload.set_chunk_size(self.chunk_size)
        if text_mode and not self.force_binary:
            tw = upload.text_writer()
            for i in iter:
                if i == self.error_at:
                    raise MyException(f"Oops {i}")
                if upload.is_cancelled() and not self.ignore_cancel:
                    self.cancelled_at = i
                    break
                s = f"{i}\n"
                tw.write(s)
        else:
            bw = upload.binary_writer()
            for i in iter:
                if i == self.error_at:
                    raise MyException(f"Oops {i}")
                if upload.is_cancelled() and not self.ignore_cancel:
                    self.cancelled_at = i
                    break
                s = f"{i}\n"
                bw.write(bytes(s, 'ascii'))


class MyDownloader(Downloader):
    lines: Optional[int] = None
    error_at_line: Optional[int] = None
    refuse: Optional[str] = None
    forget_to_return_after_refusal: bool = False
    buffer: StringIO

    def __init__(self):
        self.buffer = StringIO()

    def handle_download(self, download: Download, filename: str, text_mode: bool):
        if self.refuse:
            download.send_error(self.refuse)
            if not self.forget_to_return_after_refusal:
                return
        if self.lines is None:
            copyfileobj(download.text_reader(), self.buffer)
        else:
            tr = download.text_reader()
            i = 0
            while self.lines is None or self.lines < 0 or i < self.lines:
                if i == self.error_at_line:
                    raise MyException("oopsie")
                else:
                    line = tr.readline()
                    if not line:
                        break
                    self.buffer.write(line)
                i += 1

    def get(self):
        return self.buffer.getvalue()


class DeadManHandle:
    """
    Watchdog timer. Kills the process after a certain time. Great to detect
    deadlocks, but can be inconvenient when running tests in the debugger. In
    that case, temporarily call deadman.cancel() at the start of the test.
    """
    def __init__(self):
        self.cond = Condition()
        self.deadline = None
        self.message = None
        self.thread = Thread(target=self.work, daemon=True)
        self.thread.start()

    def set_timeout(self, t, msg):
        self.set_deadline(time.time() + t, msg)

    def set_deadline(self, d, msg):
        with self.cond:
            self.deadline = d
            if msg:
                self.message = msg
            self.cond.notify_all()

    def cancel(self):
        with self.cond:
            self.deadline = None
            self.message = None

    def work(self):
        with self.cond:
            while True:
                now = time.time()
                if self.deadline:
                    delta = self.deadline - now
                    if delta <= 0:
                        print("\n\nTIMEOUT:", self.message, "\n\n", file=sys.stderr)
                        os.kill(os.getpid(), signal.SIGKILL)
                else:
                    delta = None
                self.cond.wait(timeout=delta)


deadman = DeadManHandle()


class Common:
    first = True
    tmpdir: Optional[Path] = None

    defaultencoding = None

    def file(self, filename):
        """Resolve the given relative path within our temp directory."""
        if not self.tmpdir:
            self.tmpdir = Path(mkdtemp(prefix="filetrans_"))
        return self.tmpdir.joinpath(filename)

    def open(self, filename, mode, **kwargs):
        """Open the given filename, resolved within our temp directory"""
        fullname = self.file(filename)
        return open(fullname, mode, **kwargs)

    def commonSetUp(self):
        with self.open('checkencoding.txt', 'wt') as f:
            self.defaultencoding = f.encoding

        self.conn = conn = connect(**test_args)
        self.uploader = MyUploader()
        conn.set_uploader(self.uploader)
        self.downloader = MyDownloader()
        conn.set_downloader(self.downloader)

        self.cursor = c = self.conn.cursor()
        if self.first:
            c.execute('DROP TABLE IF EXISTS foo')
            c.execute('CREATE TABLE foo(i INT)')
            c.execute('DROP TABLE IF EXISTS foo2')
            c.execute('CREATE TABLE foo2(i INT, t VARCHAR(20))')
            conn.commit()
            self.first = False

        deadman.set_timeout(10, f"timeout in {self._testMethodName}()")

    def commonTearDown(self):
        deadman.cancel()
        try:
            self.cursor.close()
            self.conn.rollback()
            self.conn.close()
        except MonetError:
            pass

    def fill_foo(self, nrows):
        self.execute("INSERT INTO foo(i) SELECT * FROM sys.generate_series(1, %s + 1)", [nrows])

    def execute(self, *args, **kwargs):
        return self.cursor.execute(*args, **kwargs)

    def expect(self, expected_resultset):
        actual_resultset = self.cursor.fetchall()
        self.assertEqual(expected_resultset, actual_resultset)

    def expect1(self, value):
        self.expect([(value,)])

    def compression_prefix(self, scheme):
        return {'gz': b'\x1F\x8B', 'bz2': b'\x42\x5A\x68', 'xz': b'\xFD\x37\x7A\x58\x5A\x00',
                'lz4': b'\x04\x22\x4D\x18', None: None}[scheme]


class TestFileTransfer(TestCase, Common):

    def setUp(self):
        super().setUp()
        self.commonSetUp()

    def tearDown(self):
        self.commonTearDown()
        super().tearDown()

    def test_do_nothing_at_all(self):
        self.uploader.do_nothing_at_all = True
        with self.assertRaises(ProgrammingError):
            # Handler must either refuse or create a writer.
            # Not writing to the writer is not a problem, that's just an empty file
            self.execute("COPY INTO foo FROM 'foo' ON CLIENT")

    def test_upload(self):
        self.execute("COPY INTO foo FROM 'foo' ON CLIENT")
        self.execute("SELECT COUNT(*) FROM foo")
        self.expect1(self.uploader.rows)

    def test_upload_empty(self):
        self.uploader.rows = 0
        self.execute("COPY INTO foo FROM 'foo' ON CLIENT")
        self.execute("SELECT COUNT(*) FROM foo")
        self.expect1(self.uploader.rows)

    # Also see test_NormalizeCrLf from the Java tests
    def test_upload_crlf(self):
        class CustomUploader(Uploader):
            def handle_upload(self, upload: Upload, filename: str, text_mode: bool, skip_amount: int):
                w = upload.text_writer()
                w.write("1|A\r\n2|BB\r")
                w.flush()
                w.write("\n3|CCC\r\n")

        self.conn.set_uploader(CustomUploader())
        self.execute("COPY INTO foo2 FROM 'foo2' ON CLIENT")
        self.execute("SELECT i, t FROM foo2")
        self.expect([(1, "A"), (2, "BB"), (3, "CCC")])

    def test_client_refuses_upload(self):
        # our Uploader refuses filename that start with 'x'
        with self.assertRaises(OperationalError):
            self.execute("COPY INTO foo FROM 'xfoo' ON CLIENT")

    def test_upload_offset0(self):
        # OFFSET 0 and OFFSET 1 behave the same, they do nothing
        self.uploader.chunk_size = 100
        self.uploader.rows = 100
        self.execute("COPY OFFSET 0 INTO foo FROM 'foo' ON CLIENT")
        self.execute("SELECT MIN(i) AS mi, MAX(i) AS ma FROM foo")
        self.expect([(1, 100)])

    def test_upload_offset1(self):
        # OFFSET 0 and OFFSET 1 behave the same, they do nothing
        self.uploader.chunk_size = 100
        self.uploader.rows = 100
        self.execute("COPY OFFSET 1 INTO foo FROM 'foo' ON CLIENT")
        self.execute("SELECT MIN(i) AS mi, MAX(i) AS ma FROM foo")
        self.expect([(1, 100)])

    def test_upload_offset5(self):
        self.uploader.chunk_size = 100
        self.uploader.rows = 100
        self.execute("COPY OFFSET 5 INTO foo FROM 'foo' ON CLIENT")
        self.execute("SELECT MIN(i) AS mi, MAX(i) AS ma FROM foo")
        self.expect([(5, 100)])

    def test_server_cancels_upload(self):
        # self.uploader.chunkSize = 100
        self.execute("COPY 10 RECORDS INTO foo FROM 'foo' ON CLIENT")
        self.assertGreater(self.uploader.cancelled_at, 0)
        self.execute("SELECT COUNT(*) FROM foo")
        self.expect1(10)

    def test_download_refused(self):
        self.downloader.refuse = 'no thanks'
        with self.assertRaises(OperationalError):
            self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        # connection still alive
        self.conn.rollback()
        self.execute("SELECT 42")
        self.expect1(42)

    def test_download(self):
        self.fill_foo(5)
        self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        self.assertEqual("1\n2\n3\n4\n5\n", self.downloader.get())

    def test_download_empty(self):
        self.fill_foo(0)
        self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        self.assertEqual("", self.downloader.get())

    def test_download_lines(self):
        self.downloader.lines = -1
        self.fill_foo(5)
        self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        self.assertEqual("1\n2\n3\n4\n5\n", self.downloader.get())

    def test_download_stop_reading_halfway(self):
        self.fill_foo(10000)
        self.downloader.lines = 10
        self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        self.assertEqual("1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n", self.downloader.get())
        # connection still alive
        self.execute("SELECT 42")
        self.expect1(42)

    @skipUnless(test_full, "full test disabled")
    def test_large_upload(self):
        deadman.set_timeout(50, None)
        n = 1_000_000
        self.uploader.rows = n
        self.uploader.chunk_size = 1024 * 1024
        self.execute("COPY INTO foo FROM 'foo' ON CLIENT")
        self.assertIsNone(self.uploader.cancelled_at)
        self.execute("SELECT COUNT(*) FROM foo")
        self.expect1(n)

    @skipUnless(test_full, "full test disabled")
    def test_large_download(self):
        deadman.set_timeout(50, None)
        n = 1_000_000
        self.fill_foo(n)
        self.execute("COPY (SELECT * FROM foo) INTO 'banana' ON CLIENT")
        content = self.downloader.get()
        nlines = len(content.splitlines())
        self.assertEqual(n, nlines)

    def test_upload_native_text_file(self):
        self.upload_file('native.csv', {}, True)

    def test_upload_unix_text_file(self):
        self.upload_file('unix.csv', dict(newline="\n"), True)

    def test_upload_dos_text_file(self):
        self.upload_file('dos.csv', dict(newline="\r\n"), True)

    def test_upload_unix_binary_file(self):
        self.upload_file('unix.csv', dict(newline="\n"), False)

    def upload_file(self, filename, write_opts, read_text):
        encoding = self.defaultencoding if read_text else 'utf-8'
        interesting_text = "Únïçøðε¡!÷"
        encodable_text = ""
        for c in interesting_text:
            try:
                # Not all of the following characters may be available in the system encoding
                bytes(c, self.defaultencoding)
            except UnicodeEncodeError:
                continue
            encodable_text += c
        assert len(encodable_text) > 0
        n = 1000
        f = self.open(filename, 'wt', encoding=encoding, **write_opts)
        assert f.encoding
        for i in range(n):
            print(f"{i}|{encodable_text}{i}", file=f)
        f.close()
        testcase = self

        class CustomUploader(Uploader):
            def handle_upload(self, upload: Upload, filename: str, text_mode: bool, skip_amount: int):
                if read_text:
                    f = testcase.open(filename, 'r')
                    tw = upload.text_writer()
                    copyfileobj(f, tw)
                else:
                    f = testcase.open(filename, 'rb')
                    bw = upload.binary_writer()
                    copyfileobj(f, bw)
                f.close()

        self.conn.set_uploader(CustomUploader())
        self.execute(f"COPY INTO foo2 FROM '{filename}' ON CLIENT")
        self.execute("SELECT t FROM foo2 where i = 999")
        self.expect1(f"{encodable_text}999")

    def test_fail_upload_late(self):
        self.uploader.error_at = 99
        # If the handler raises an exception, ..
        with self.assertRaises(MyException):
            self.execute("COPY INTO foo FROM 'foo' ON CLIENT")
        # .. the connection is dropped
        with self.assertRaisesRegex(ProgrammingError, "ot connected"):
            self.execute("SELECT COUNT(*) FROM foo")

    def test_download_immediate_exception(self):

        class CustomDownloader(Downloader):
            def handle_download(self, download: Download, filename: str, text_mode: bool):
                raise MyException("fail early")

        self.fill_foo(5000)
        self.conn.set_downloader(CustomDownloader())
        # If the handler raises an exception, ..
        with self.assertRaises(MyException):
            self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        # .. the connection is dropped
        with self.assertRaisesRegex(ProgrammingError, "ot connected"):
            self.execute("SELECT COUNT(*) FROM foo")

    def test_fail_download_late(self):
        self.fill_foo(5000)
        self.downloader.lines = 6000
        self.downloader.error_at_line = 4000
        # If the handler raises an exception, ..
        with self.assertRaises(MyException):
            self.execute("COPY (SELECT * FROM foo) INTO 'foo' ON CLIENT")
        # .. the connection is dropped
        with self.assertRaisesRegex(ProgrammingError, "ot connected"):
            self.execute("SELECT COUNT(*) FROM foo")


class TestSafeDirectoryHandler(TestCase, Common):

    def setUp(self):
        super().setUp()
        self.commonSetUp()

    def tearDown(self):
        self.commonTearDown()
        super().tearDown()

    def test_upload_handler_security(self):
        f = self.open("foo.csv", "w")
        f.write("1\n2\n3\n")
        f.close()
        outside = self.file('')
        inside = self.file('inside')
        inside.mkdir()
        f = self.open(inside.joinpath("foo.csv"), "w")
        f.write("10\n20\n30\n")
        f.close()
        #
        handler = SafeDirectoryHandler(inside)
        self.conn.set_uploader(handler)
        #
        testcases = [
            ('foo.csv', True),
            ('./foo.csv', True),
            (inside.joinpath('foo.csv'), True),
            ('../foo.csv', False),
            (outside.joinpath('foo.csv'), False),
        ]
        for path, valid in testcases:
            with self.subTest(dir=str(inside), path=str(path), expect_valid=valid):
                self.conn.rollback()
                path = str(path)
                if valid:
                    self.execute("COPY INTO foo FROM %s ON CLIENT", [path])
                else:
                    with self.assertRaises(OperationalError):
                        self.execute("COPY INTO foo FROM %s ON CLIENT", [path])
                    continue
                self.execute("SELECT MAX(i) FROM foo")
                self.expect1(30)

    def test_download_handler_security(self):
        self.execute("INSERT INTO foo SELECT * FROM sys.generate_series(0, 10)")
        outside = self.file('')
        inside = self.file('inside')
        inside.mkdir()
        #
        handler = SafeDirectoryHandler(inside)
        self.conn.set_downloader(handler)
        #
        testcases = [
            ('foo.csv', True),
            ('./foo.csv', True),
            (inside.joinpath('foo.csv'), True),
            ('../foo.csv', False),
            (outside.joinpath('foo.csv'), False),
        ]
        for path, valid in testcases:
            with self.subTest(dir=str(inside), path=str(path), expect_valid=valid):
                self.conn.rollback()
                path = str(path)
                if valid:
                    self.execute("COPY (SELECT * FROM foo) INTO %s ON CLIENT", [path])
                else:
                    with self.assertRaises(OperationalError):
                        self.execute("COPY (SELECT * FROM foo) INTO %s ON CLIENT", [path])
                    continue

    def get_testdata_name(self, enc_name: str, newline: str, lines: int = None, compression=None) -> str:
        newline_name = {None: "none", "\n": "lf", "\r\n": "crlf"}[newline]
        file_name = f"{enc_name}_{newline_name}"
        if lines is not None:
            file_name += f"_{lines}lines"
        file_name += ".txt"
        if compression:
            file_name += "." + compression
        return file_name

    def get_testdata(self, enc_name: str, newline: str, lines: int, compression: str = None) -> str:
        encoding = codecs.lookup(enc_name) if enc_name else None
        fname = self.get_testdata_name(enc_name, newline, lines, compression)
        p = self.file(fname)
        if not p.exists():
            enc = encoding.name if encoding else None
            opener = lookup_compression_algorithm(p)
            f = opener(p, mode="wt", encoding=enc, newline=newline)
            for n in range(lines):
                i, t = self.line(n)
                print(f"{i}|{t}", file=f)
            f.close()
            assert p.exists()

        return fname

    def line(self, i: int) -> Tuple[int, str]:
        k = i + 1
        if i % 7 == 0:
            s = ""
        else:
            # ÷ is interesting because it appears in all of UTF-8, Latin1 and
            # Shift-JIS, but with different encodings.
            s = f"÷{k}"
        return (k, s)

    def test_utf8_uploads(self):
        self.perform_line_endings_and_offsets_upload_tests('utf-8')

    def test_latin1_uploads(self):
        self.perform_line_endings_and_offsets_upload_tests('latin1')

    def test_shiftjis_uploads(self):
        self.perform_line_endings_and_offsets_upload_tests('shift-jis')

    def test_native_uploads(self):
        self.perform_line_endings_and_offsets_upload_tests(None)

    def perform_line_endings_and_offsets_upload_tests(self, enc):
        endings = [
            ('\n', '\n'),
            ('\r\n', '\r\n'),
            ('\n', None),
            ('\r\n', None),
        ]
        offsets = {
            None,
            0,
            1,
            2,
            5,
            15,
        }
        for file_ending, handler_ending in endings:
            for offset in offsets:
                with self.subTest(encoding=enc, file_ending=file_ending, handler_ending=handler_ending, offset=offset):
                    self.perform_upload_test(enc, file_ending, handler_ending, offset)

    def perform_upload_test(self, encoding, file_ending, handler_ending, offset=None, end=10, compression=None):
        if offset is None:
            offset_clause = ''
            skip = 0
        else:
            offset_clause = f" OFFSET {offset}"
            skip = offset - 1 if offset else 0
        uploader = SafeDirectoryHandler(self.file(''), encoding, handler_ending)
        self.conn.set_uploader(uploader)
        fname = self.get_testdata(encoding, file_ending, end, compression=compression)
        # Double check the compression, are we testing what we want tot test?
        compression_prefix = self.compression_prefix(compression)
        if compression_prefix:
            f = self.open(fname, 'rb')
            content = f.read()
            content_prefix = content[:len(compression_prefix)]
            f.close()
            self.assertEqual(compression_prefix, content_prefix)
        # Double check the testdata encoding, are we testing what we want tot test?
        # These are the various encodings of the '÷' character as used by the
        # .line() method above.
        encmarker = {'utf-8': b'\xC3\xB7', 'latin1': b'\xF7', 'shift-jis': b'\x81\x80', None: None}[encoding]
        if encmarker:
            full_name = self.file(fname)
            opener = lookup_compression_algorithm(full_name)
            f = opener(full_name, 'rb')
            content = f.read()
            f.close()
            self.assertTrue(content == b'' or encmarker in content)
        # Run the test
        # self.conn.rollback()
        self.execute("DELETE FROM foo2")
        self.execute("COPY" + offset_clause + " INTO foo2 FROM %s ON CLIENT", [fname])
        self.execute("SELECT * FROM foo2")
        rows = self.cursor.fetchall()
        expected = [self.line(i) for i in range(skip, end)]
        self.assertEqual(expected, rows)

    def test_upload_utf8_lf_uses_binary(self):
        class CustomHandler(SafeDirectoryHandler):
            used_mode = None

            def __init__(self, dir):
                super().__init__(dir, 'utf-8', '\n')

            def handle_upload(self, upload: Upload, filename: str, text_mode: bool, skip_amount: int):
                super().handle_upload(upload, filename, text_mode, skip_amount)
                # peek into the internals of the upload to see what was used.
                if upload.writer:
                    self.used_mode = 'binary'
                if upload.twriter:
                    # overwrite
                    self.used_mode = 'text'

        fname = self.get_testdata('utf-8', '\n', 10)
        uploader = CustomHandler(self.file(''))
        self.conn.set_uploader(uploader)
        self.execute("COPY INTO foo2 FROM %s ON CLIENT", fname)
        self.assertEqual('binary', uploader.used_mode)

    def test_download_encodings_and_line_endings(self):
        compressions = [
            None,
            'gz',
            'bz2',
            'xz',
        ]
        if have_lz4:
            compressions.append('lz4')
        encodings = [
            'utf-8',
            'latin1',
            'shift-jis',
            None  # means native
        ]
        file_endings = [
            "\n",
            "\r\n",
            None,
        ]
        for compression in compressions:
            for encoding in encodings:
                for handler_ending in file_endings + [None]:
                    with self.subTest(compression=compression, encoding=encoding, handler_ending=handler_ending):
                        self.perform_download_test(compression, encoding, handler_ending)

    def perform_download_test(self, compression, encoding, handler_ending):
        # We want to check that when asked to use the given encoding and line endings,
        # this happens.
        n = 10
        downloader = SafeDirectoryHandler(self.file(''), encoding, handler_ending)
        self.conn.set_downloader(downloader)
        self.execute("DELETE FROM foo2")
        #
        enc = encoding or self.defaultencoding
        expected = b""
        for k in range(n):
            i, s = self.line(k)
            expected += bytes(str(i), enc)
            expected += b'|"'
            expected += bytes(s, enc)
            expected += b'"'
            expected += bytes(handler_ending or os.linesep, enc)
            self.execute("INSERT INTO foo2(i, t) VALUES (%s, %s)", [i, s])
        #
        fname = self.get_testdata_name(encoding, handler_ending, compression=compression)
        self.execute("COPY (SELECT * FROM foo2) INTO %s ON CLIENT", [fname])
        # check compression
        compression_prefix = self.compression_prefix(compression)
        if compression_prefix:
            f = self.open(fname, 'rb')
            content = f.read()
            content_prefix = content[:len(compression_prefix)]
            f.close()
            self.assertEqual(compression_prefix, content_prefix)
        # check contents
        full_name = self.file(fname)
        opener = lookup_compression_algorithm(full_name)
        f = opener(full_name, 'rb')
        content = f.read()
        f.close()
        #
        self.assertEqual(expected, content)

    def test_download_utf8_lf_uses_binary(self):
        class CustomHandler(SafeDirectoryHandler):
            used_mode = None

            def __init__(self, dir):
                super().__init__(dir, 'utf-8', '\n')

            def handle_download(self, download: Download, filename: str, text_mode: bool):
                super().handle_download(download, filename, text_mode)
                # peek into the internals of the download
                if download.reader:
                    self.used_mode = 'binary'
                if download.treader:
                    # overwrite
                    self.used_mode = 'text'

        fname = self.get_testdata_name('utf-8', '\n')
        downloader = CustomHandler(self.file(''))
        self.conn.set_downloader(downloader)
        self.execute("COPY SELECT * FROM sys.generate_series(0,10) INTO %s ON CLIENT", fname)
        self.assertEqual('binary', downloader.used_mode)

    def test_upload_with_compression_disabled(self):
        fname = self.get_testdata('utf-8', '\n', 3, compression=None)
        # give it a misleading name
        misleading_name = 'banana.txt.gz'
        copyfile(self.file(fname), self.file(misleading_name))
        # now upload it
        handler = SafeDirectoryHandler(self.file(''), compression=False)
        self.conn.set_uploader(handler)
        self.execute("COPY INTO foo2 FROM %s ON CLIENT", misleading_name)
        self.conn.commit()
        self.execute("SELECT MAX(i) FROM foo2")
        self.expect1(3)

    def test_download_with_compression_disabled(self):
        fname = 'misleading.txt.gz'
        handler = SafeDirectoryHandler(self.file(''), encoding='utf-8', newline='\n', compression=False)
        self.conn.set_downloader(handler)
        self.execute("COPY SELECT value FROM sys.generate_series(1,4) INTO %s ON CLIENT", fname)
        # should not be gzipped
        f = self.open(fname, 'rb')
        content = f.read()
        f.close()
        self.assertEqual(b'1\n2\n3\n', content)


class TestNormalizeCrLf(TestCase):

    class Sink(BufferedIOBase):
        def __init__(self):
            self.written = b''

        def writable(self) -> bool:
            return True

        def write(self, buf):
            self.written += bytes(buf)
            return len(buf)

        def get_written(self):
            res = self.written
            self.written = b''
            return res

    def setUp(self):
        self.sink = self.Sink()
        self.normalizer = NormalizeCrLf(self.sink)

    def transaction(self, buf, expect_pending, expect_written):
        n = self.normalizer.write(buf)
        self.assertEqual(len(buf), n)
        written = self.sink.get_written()
        pending = self.normalizer.pending
        self.assertEqual(expect_written, written)
        self.assertEqual(expect_pending, pending)

    def test_normalizer(self):
        self.assertEqual(False, self.normalizer.pending)
        self.assertEqual(b'', self.sink.written)

        # can all be written through
        self.transaction(b"\r\naaa\n\n\r\n", False, b"\naaa\n\n\n")

        # trailing CR pending
        self.transaction(b"\n\r\naaa\r", True, b"\n\naaa")

        # LF consumes the pending CR
        self.transaction(b"\n", False, b"\n")

        # a new pending CR
        self.transaction(b"\r", True, b"")

        # a new pending CR
        self.transaction(b"a", False, b"\ra")

        # CR after CR emits one CR and stays pending
        self.transaction(b"\r", True, b"")
        self.transaction(b"\r", True, b"\r")

        # empty write stays pending
        self.transaction(b"", True, b"")

        # flushing the normalizer does not flush the pending CR
        self.normalizer.flush()
        self.assertTrue(self.normalizer.pending)

        # but closing it does
        self.normalizer.close()
        self.assertFalse(self.normalizer.pending)
        self.assertEqual(b"\r", self.sink.get_written())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

"""
functions for converting python objects to monetdb SQL format. If you want
to add support for a specific type you should add a function as a value to
the mapping dict and the datatype as key.
"""

import datetime
import decimal
import uuid

from pymonetdb.exceptions import ProgrammingError


def monet_none(_):
    """
    returns a NULL string
    """
    return "NULL"


def monet_bool(data):
    """
    returns "true" or "false"
    """
    return ["false", "true"][bool(data)]


def monet_escape(data):
    """
    returns an escaped string
    """
    data = str(data).replace("\\", "\\\\")
    data = data.replace("\'", "\\\'")
    return "'%s'" % str(data)


def monet_bytes(data):
    """
    converts bytes to string
    """
    return "BLOB '%s'" % data.hex()


def _tzaware(time_or_datetime):
    """
    returns True if the time or datetime is timezone aware, False otherwise.
    """
    return time_or_datetime.utcoffset() is not None


def monet_datetime(data):
    """
    returns a casted timestamp
    """
    if _tzaware(data):
        return "TIMESTAMPTZ %s" % monet_escape(data)
    else:
        return "TIMESTAMP %s" % monet_escape(data)


def monet_date(data):
    """
    returns a casted date
    """
    return "DATE %s" % monet_escape(data)


def monet_time(data):
    """
    returns a casted time
    """
    if _tzaware(data):
        return "TIMETZ %s" % monet_escape(data)
    else:
        return "TIME %s" % monet_escape(data)


def monet_timedelta(data):
    """
    returns timedelta casted to interval seconds
    """
    return "INTERVAL %s SECOND" % monet_escape(int(data.total_seconds()))


def monet_unicode(data):
    return monet_escape(data.encode('utf-8'))


mapping = [

    (str, monet_escape),
    (bytes, monet_bytes),
    (int, str),
    (complex, str),
    (float, str),
    (uuid.UUID, monet_escape),
    (decimal.Decimal, str),
    (datetime.datetime, monet_datetime),
    (datetime.time, monet_time),
    (datetime.date, monet_date),
    (datetime.timedelta, monet_timedelta),
    (bool, monet_bool),
    (type(None), monet_none),
]

mapping_dict = dict(mapping)


def convert(data):
    """
    Return the appropriate convertion function based upon the python type.
    """
    if type(data) in mapping_dict:
        return mapping_dict[type(data)](data)
    else:
        for type_, func in mapping:
            if issubclass(type(data), type_):
                return func(data)
    raise ProgrammingError("type %s not supported as value" % type(data))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import unittest
from pymonetdb.control import Control
from pymonetdb.exceptions import OperationalError
from tests.util import test_hostname, test_port, test_passphrase, test_full

database_prefix = 'controltest_'
database_name = database_prefix + 'other'


def do_without_fail(function):
    try:
        function()
    except OperationalError:
        pass


class TestControl(unittest.TestCase):
    """
    These tests require control access to monetdb. This is enabled with:

    $ monetdbd set control=yes /var/lib/monetdb
    $ monetdbd set passphrase=testdb /var/lib/monetdb

    Where /var/lib/monetdb is the path to your dbfarm. Don't forget to restart the db after setting the credentials.
    """

    def setUp(self):
        # use tcp
        self.control = Control(test_hostname, test_port, test_passphrase)

        # use socket
        # self.control = Control()

        do_without_fail(lambda: self.control.stop(database_name))
        do_without_fail(lambda: self.control.destroy(database_name))
        self.control.create(database_name)

    def _cleanup(self, old, new):
        do_without_fail(lambda: self.control.destroy(old))
        do_without_fail(lambda: self.control.destroy(new))
        self.control.create(old)
        self.control.rename(old, new)
        statuses = self.control.status()
        self.assertTrue(new in [status["name"] for status in statuses])
        do_without_fail(lambda: self.control.destroy(new))
        return statuses

    def tear_down(self):
        do_without_fail(lambda: self.control.stop(database_name))
        do_without_fail(lambda: self.control.destroy(database_name))

    def test_create(self):
        create_name = database_prefix + "create"
        do_without_fail(lambda: self.control.destroy(create_name))
        self.control.create(create_name)
        self.assertRaises(OperationalError, self.control.create, create_name)
        do_without_fail(lambda: self.control.destroy(create_name))

    def test_destroy(self):
        destroy_name = database_prefix + "destroy"
        self.control.create(destroy_name)
        self.control.destroy(destroy_name)
        self.assertRaises(OperationalError, self.control.destroy, destroy_name)

    def test_lock(self):
        do_without_fail(lambda: self.control.release(database_name))
        self.control.lock(database_name)
        with self.assertRaises(OperationalError):
            self.control.lock(database_name)
        self.control.release(database_name)

    def test_release(self):
        do_without_fail(lambda: self.control.release(database_name))
        do_without_fail(lambda: self.control.lock(database_name))
        self.assertTrue(self.control.release(database_name))
        self.assertRaises(OperationalError, self.control.release, database_name)

    @unittest.skipUnless(test_full, "full test disabled")
    def test_status(self):
        status = self.control.status(database_name)
        self.assertEqual(status["name"], database_name)

    @unittest.skipUnless(test_full, "full test disabled")
    def test_statuses(self):
        status1 = database_prefix + "status1"
        status2 = database_prefix + "status2"
        statuses = self._cleanup(status1, status2)
        self.assertTrue(status2 in [status["name"] for status in statuses])
        do_without_fail(lambda: self.control.destroy(status1))
        do_without_fail(lambda: self.control.destroy(status2))

    def test_start(self):
        do_without_fail(lambda: self.control.stop(database_name))
        self.assertTrue(self.control.start(database_name))

    @unittest.skipUnless(test_full, "full test disabled")
    def test_stop(self):
        do_without_fail(lambda: self.control.start(database_name))
        self.assertTrue(self.control.stop(database_name))

    @unittest.skipUnless(test_full, "full test disabled")
    def test_kill(self):
        do_without_fail(lambda: self.control.start(database_name))
        self.assertTrue(self.control.kill(database_name))

    def test_set(self):
        property_ = "readonly"
        value = "yes"
        self.control.set(database_name, property_, value)
        properties = self.control.get(database_name)
        self.assertEqual(properties[property_], value)

    def test_get(self):
        self.control.get(database_name)

    def test_inherit(self):
        self.control.set(database_name, "readonly", "yes")
        self.assertTrue("readonly" in self.control.get(database_name))
        self.control.inherit(database_name, "readonly")
        # TODO: False on OSX, True on travis?
        # self.assertTrue("readonly" in self.control.get(database_name))

    def test_rename(self):
        old = database_prefix + "old"
        new = database_prefix + "new"
        self._cleanup(old, new)
        do_without_fail(lambda: self.control.destroy(new))

    def test_defaults(self):
        defaults = self.control.defaults()
        self.assertTrue("readonly" in defaults)

    @unittest.skipUnless(test_full, "full test disabled")
    def test_neighbours(self):
        self.control.neighbours()



""" Script to test database capabilities and the DB-API interface
    for functionality and memory leaks.

    Adapted from a script by M-A Lemburg and taken from the MySQL python driver.

"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import datetime
from time import time
import unittest

from pymonetdb.exceptions import ProgrammingError
import pymonetdb
from pymonetdb.sql import monetize
from tests.util import test_args


class DatabaseTest(unittest.TestCase):
    create_table_extra = ''
    rows = 10

    leak_test = False

    def setUp(self):
        db = pymonetdb.connect(autocommit=False, **test_args)
        self.connection = db
        self.cursor = db.cursor()
        self.BLOBText = ''.join([chr(i) for i in range(33, 127)] * 100)
        self.BLOBBinary = bytes(range(256)) * 16
        self.BLOBUText = ''.join([chr(i) for i in range(1, 16384)])

    def tearDown(self):
        self.connection.close()

    def table_exists(self, name):
        try:
            self.cursor.execute('select * from %s where 1=0' % name)
        except pymonetdb.OperationalError:
            self.connection.rollback()
            return False
        else:
            return True

    @staticmethod
    def quote_identifier(ident):
        return '"%s"' % ident

    def new_table_name(self):
        i = id(self.cursor)
        while True:
            name = self.quote_identifier('tb%08x' % i)
            if not self.table_exists(name):
                return name
            i = i + 1

    def create_table(self, columndefs):
        """ Create a table using a list of column definitions given in
            columndefs.

            generator must be a function taking arguments (row_number,
            col_number) returning a suitable data object for insertion
            into the table.

        """
        self.table = self.new_table_name()
        self.cursor.execute('CREATE TABLE %s (%s) %s' %
                            (self.table,
                             ',\n'.join(columndefs),
                             self.create_table_extra))

    def check_data_integrity(self, columndefs, generator, drop=True):
        self.create_table(columndefs)
        insert_statement = ('INSERT INTO %s VALUES (%s)' %
                            (self.table,
                             ','.join(['%s'] * len(columndefs))))
        data = [[generator(i, j) for j in range(len(columndefs))]
                for i in range(self.rows)]
        self.cursor.executemany(insert_statement, data)
        self.connection.commit()

        self.cursor.execute('select * from %s' % self.table)
        rows = self.cursor.fetchall()
        self.assertEqual(len(rows), self.rows)
        try:
            for i in range(self.rows):
                for j in range(len(columndefs)):
                    self.assertEqual(rows[i][j], generator(i, j))
        finally:
            if drop:
                self.cursor.execute('drop table %s' % self.table)

    def test_transactions(self):
        columndefs = ('col1 INT', 'col2 VARCHAR(255)')

        def generator(row, col):
            if col == 0:
                return row
            else:
                return ('%i' % (row % 10)) * 255

        self.check_data_integrity(columndefs, generator, drop=False)
        delete_statement = 'delete from %s where col1=%%s' % self.table
        self.cursor.execute(delete_statement, (0,))
        self.cursor.execute('select col1 from %s where col1=%s' % (self.table, 0))
        rows = self.cursor.fetchall()
        self.assertFalse(rows, "DELETE didn't work")
        self.connection.rollback()
        self.cursor.execute('select col1 from %s where col1=%s' % (self.table, 0))
        rows = self.cursor.fetchall()
        self.assertTrue(len(rows) == 1, "ROLLBACK didn't work")
        self.cursor.execute('drop table %s' % self.table)

    def test_truncation(self):  # noqa: C901
        columndefs = ('col1 INT', 'col2 VARCHAR(255)')

        def generator(row, col):
            if col == 0:
                return row
            else:
                return ('%i' % (row % 10)) * (int(255 - self.rows / 2) + row)

        self.create_table(columndefs)
        insert_statement = ('INSERT INTO %s VALUES (%s)' %
                            (self.table, ','.join(['%s'] * len(columndefs))))
        try:
            self.cursor.execute(insert_statement, (0, '0' * 256))
        except Warning:
            pass
        except self.connection.Error:
            pass
        else:
            self.fail("Over-long column did not generate warnings/exception with single insert")

        self.connection.rollback()
        self.create_table(columndefs)

        try:
            for i in range(self.rows):
                data = []
                for j in range(len(columndefs)):
                    data.append(generator(i, j))
                self.cursor.execute(insert_statement, tuple(data))
        except Warning:
            pass
        except self.connection.Error:
            pass
        else:
            self.fail("Over-long columns did not generate warnings/exception with execute()")

        self.connection.rollback()
        self.create_table(columndefs)

        try:
            data = [[generator(i, j) for j in range(len(columndefs))]
                    for i in range(self.rows)]
            self.cursor.executemany(insert_statement, data)
        except Warning:
            pass
        except self.connection.Error:
            pass
        else:
            self.fail("Over-long columns did not generate warnings/exception with executemany()")

        self.connection.rollback()

    def test_CHAR(self):
        # Character data
        def generator(row, col):
            return ('%i' % ((row + col) % 10)) * 255

        self.check_data_integrity(
            ('col1 char(255)', 'col2 char(255)'),
            generator)

    def test_INT(self):
        # Number data
        def generator(row, _):
            return row * row

        self.check_data_integrity(
            ('col1 INT',),
            generator)

    def test_DECIMAL(self):
        def generator(row, col):
            from decimal import Decimal
            return Decimal("%d.%02d" % (row, col))

        self.check_data_integrity(
            ('col1 DECIMAL(5,2)',),
            generator)

    def test_REAL(self):
        def generator(row, _):
            return row * 1000.0

        self.check_data_integrity(
            ('col1 REAL',),
            generator)

    def test_DOUBLE(self):
        def generator(row, _):
            return row / 1e-99

        self.check_data_integrity(
            ('col1 DOUBLE',),
            generator)

    def test_DATE(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.DateFromTicks(ticks + row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 DATE',),
            generator)

    def test_TIME(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.TimeFromTicks(ticks + row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 TIME',),
            generator)

    def test_TIMETZ(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.TimeTzFromTicks(ticks + row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 TIMETZ',),
            generator)

    def test_DATETIME(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.TimestampFromTicks(ticks + row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 TIMESTAMP',),
            generator)

    def test_TIMESTAMP(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.TimestampFromTicks(ticks + row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 TIMESTAMP',),
            generator)

    def test_TIMESTAMPTZ(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.TimestampTzFromTicks(ticks + row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 TIMESTAMPTZ',),
            generator)

    def test_fractional_TIMESTAMP(self):
        ticks = time()

        def generator(row, col):
            return pymonetdb.TimestampFromTicks(ticks + row * 86400 - col * 1313 + row * 0.7 * col / 3.0)

        self.check_data_integrity(
            ('col1 TIMESTAMP',),
            generator)

    def test_SEC_INTERVAL(self):

        def generator(row, col):
            return datetime.timedelta(seconds=row * 86400 - col * 1313)

        self.check_data_integrity(
            ('col1 INTERVAL SECOND',),
            generator)

    def test_TEXT(self):
        def generator(_, __):
            return self.BLOBText  # 'BLOB Text ' * 1024

        self.check_data_integrity(
            ('col2 TEXT',),
            generator)

    def test_BLOB(self):
        def generator(row, col):
            if col == 0:
                return row
            else:
                return self.BLOBBinary  # 'BLOB\000Binary ' * 1024

        self.check_data_integrity(
            ('col1 INT', 'col2 BLOB'),
            generator)

    def test_TINYINT(self):
        # Number data
        def generator(row, _):
            v = (row * row) % 256
            if v > 127:
                v = v - 256
            return v

        self.check_data_integrity(
            ('col1 TINYINT',),
            generator)

    def test_small_CHAR(self):
        # Character data
        def generator(row, col):
            i = (row * col + 62) % 256
            if i == 62:
                return ''
            if i == 63:
                return None
            return chr(i)

        self.check_data_integrity(
            ('col1 char(1)', 'col2 char(1)'),
            generator)

    def test_BOOL(self):
        def generator(row, _):
            return bool(row % 2)

        self.check_data_integrity(
            ('col1 BOOL',),
            generator)

    def test_URL(self):
        def generator(_, __):
            return "http://example.org/something"

        self.check_data_integrity(
            ('col1 URL',),
            generator)

    def test_INET(self):
        def generator(_, __):
            return "192.168.254.101"

        self.check_data_integrity(
            ('col1 INET',),
            generator)

    def test_description(self):
        self.table = self.new_table_name()
        shouldbe = [
            ('c', 'varchar', None, 1024, None, None, None),
            ('d', 'decimal', None, 9, 9, 4, None),
            ('n', 'varchar', None, 1, None, None, None),
        ]
        try:
            self.cursor.execute(
                "create table %s (c VARCHAR(1024), d DECIMAL(9,4), n VARCHAR(1) NOT NULL)" % self.table)
            self.cursor.execute("insert into %s VALUES ('test', 12345.1234, 'x')" % self.table)
            self.cursor.execute('select * from %s' % self.table)
            self.assertEqual(self.cursor.description, shouldbe, "cursor.description is incorrect")
        finally:
            self.cursor.execute('drop table %s' % self.table)

    def test_bigresult(self):
        self.cursor.execute('select count(*) from types')
        r = self.cursor.fetchone()
        n = r[0]
        self.cursor.arraysize = 100000
        self.cursor.execute('select * from types t1, types t2')
        r = self.cursor.fetchall()
        self.assertEqual(len(r), n ** 2)

    def test_closecur(self):
        self.cursor.close()
        self.assertRaises(ProgrammingError, self.cursor.execute, "select * from tables")
        self.cursor = self.connection.cursor()

    def test_customtype(self):
        t = ["list", "test"]
        self.assertRaises(ProgrammingError, monetize.convert, t)
        monetize.mapping_dict[list] = str
        self.assertEqual(monetize.convert(t), "['list', 'test']")

    def test_multiple_queries(self):
        table1 = self.new_table_name()
        table2 = table1[:-1] + 'bla"'
        self.cursor.execute("create table %s (a int)" % table1)
        self.cursor.execute("create table %s (a int, b int)" % table2)
        self.cursor.execute("insert into %s VALUES (100)" % table1)
        self.cursor.execute("insert into %s VALUES (50, 50)" % table2)
        self.cursor.execute('select * from %s; select * from %s;' %
                            (table1, table2))
        result = self.cursor.fetchall()
        self.assertEqual(result, [(50, 50)])

    def test_temporal_operations(self):
        dt = datetime.datetime(2017, 12, 6, 12, 30)
        self.cursor.execute("SELECT %(dt)s - INTERVAL '1' DAY", {'dt': dt})
        expected = datetime.datetime(2017, 12, 5, 12, 30)
        self.assertEqual(self.cursor.fetchone()[0], expected)

        d = datetime.date(2017, 12, 6)
        self.cursor.execute("SELECT %(d)s - INTERVAL '1' DAY", {'d': d})
        expected = datetime.date(2017, 12, 5)
        self.assertEqual(self.cursor.fetchone()[0], expected)

        t = datetime.time(12, 5)
        self.cursor.execute("SELECT %(t)s - INTERVAL '30' MINUTE", {'t': t})
        expected = datetime.time(11, 35)
        self.assertEqual(self.cursor.fetchone()[0], expected)

        td = datetime.timedelta(days=5, hours=2, minutes=10)
        self.cursor.execute("SELECT %(dt)s - %(td)s", {'dt': dt, 'td': td})
        expected = datetime.datetime(2017, 12, 1, 10, 20)
        self.assertEqual(self.cursor.fetchone()[0], expected)



"""
Python DB API 2.0 driver compliance unit test suite.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
#
#
# this code is modified version of (public domain):
#
# https://github.com/psycopg/psycopg2/blob/master/tests/dbapi20.py
#
import unittest
import time
import pymonetdb
from tests.util import test_args


class DatabaseAPI20Test(unittest.TestCase):
    """ Test pymonetdb for DB API 2.0 compatibility.
    """

    table_prefix = 'dbapi20test_'  # If you need to specify a prefix for tables

    ddl1 = 'create table %sbooze (name varchar(20))' % table_prefix
    ddl2 = 'create table %sbarflys (name varchar(20))' % table_prefix
    xddl1 = 'drop table %sbooze' % table_prefix
    xddl2 = 'drop table %sbarflys' % table_prefix

    lowerfunc = 'lower'  # Name of stored procedure to convert string->lowercase

    # Some drivers may need to override these helpers, for example adding
    # a 'commit' after the execute.
    def executeDDL1(self, cursor):
        cursor.execute(self.ddl1)

    def executeDDL2(self, cursor):
        cursor.execute(self.ddl2)

    def setUp(self):
        pass

    def tearDown(self):
        con = self._connect()
        try:
            cur = con.cursor()
            for ddl in (self.xddl1, self.xddl2):
                try:
                    cur.execute(ddl)
                    con.commit()
                except pymonetdb.Error:
                    # Assume table didn't exist. Other tests will check if
                    # execute is busted.
                    pass
        finally:
            con.close()

    def _connect(self):
        try:
            return pymonetdb.connect(**test_args)
        except AttributeError:
            self.fail("No connect method found in pymonetdb module")

    def test_connect(self):
        con = self._connect()
        con.close()

    def test_apilevel(self):
        try:
            # Must exist
            apilevel = pymonetdb.apilevel
            # Must equal 2.0
            self.assertEqual(apilevel, '2.0')
        except AttributeError:
            self.fail("Driver doesn't define apilevel")

    def test_threadsafety(self):
        try:
            # Must exist
            threadsafety = pymonetdb.threadsafety
            # Must be a valid value
            self.assertTrue(threadsafety in (0, 1, 2, 3))
        except AttributeError:
            self.fail("Driver doesn't define threadsafety")

    def test_paramstyle(self):
        try:
            # Must exist
            paramstyle = pymonetdb.paramstyle
            # Must be a valid value
            self.assertTrue(paramstyle in (
                'qmark', 'numeric', 'named', 'format', 'pyformat'
            ))
        except AttributeError:
            self.fail("Driver doesn't define paramstyle")

    def test_Exceptions(self):
        # Make sure required exceptions exist, and are in the
        # defined hierarchy.
        # under Python 3 StardardError no longer exist, replaced with Exception
        self.assertTrue(issubclass(pymonetdb.Warning, Exception))
        self.assertTrue(issubclass(pymonetdb.Error, Exception))

        self.assertTrue(
            issubclass(pymonetdb.InterfaceError, pymonetdb.Error)
        )
        self.assertTrue(
            issubclass(pymonetdb.DatabaseError, pymonetdb.Error)
        )
        self.assertTrue(
            issubclass(pymonetdb.OperationalError, pymonetdb.Error)
        )
        self.assertTrue(
            issubclass(pymonetdb.IntegrityError, pymonetdb.Error)
        )
        self.assertTrue(
            issubclass(pymonetdb.InternalError, pymonetdb.Error)
        )
        self.assertTrue(
            issubclass(pymonetdb.ProgrammingError, pymonetdb.Error)
        )
        self.assertTrue(
            issubclass(pymonetdb.NotSupportedError, pymonetdb.Error)
        )

    def test_ExceptionsAsConnectionAttributes(self):
        # OPTIONAL EXTENSION
        # Test for the optional DB API 2.0 extension, where the exceptions
        # are exposed as attributes on the Connection object
        # I figure this optional extension will be implemented by any
        # driver author who is using this test suite, so it is enabled
        # by default.
        con = self._connect()
        drv = pymonetdb
        self.assertTrue(con.Warning is drv.Warning)
        self.assertTrue(con.Error is drv.Error)
        self.assertTrue(con.InterfaceError is drv.InterfaceError)
        self.assertTrue(con.DatabaseError is drv.DatabaseError)
        self.assertTrue(con.OperationalError is drv.OperationalError)
        self.assertTrue(con.IntegrityError is drv.IntegrityError)
        self.assertTrue(con.InternalError is drv.InternalError)
        self.assertTrue(con.ProgrammingError is drv.ProgrammingError)
        self.assertTrue(con.NotSupportedError is drv.NotSupportedError)

    def test_commit(self):
        con = self._connect()
        try:
            # Commit must work, even if it doesn't do anything
            con.commit()
        finally:
            con.close()

    def test_rollback(self):
        con = self._connect()
        # If rollback is defined, it should either work or throw
        # the documented exception
        if hasattr(con, 'rollback'):
            try:
                con.rollback()
            except pymonetdb.NotSupportedError:
                pass

    def test_cursor(self):
        con = self._connect()
        try:
            _ = con.cursor()
        finally:
            con.close()

    def test_cursor_isolation(self):
        con = self._connect()
        try:
            # Make sure cursors created from the same connection have
            # the documented transaction isolation level
            cur1 = con.cursor()
            cur2 = con.cursor()
            self.executeDDL1(cur1)
            cur1.execute("insert into %sbooze values ('Victoria Bitter')" % (
                self.table_prefix
            ))
            cur2.execute("select name from %sbooze" % self.table_prefix)
            booze = cur2.fetchall()
            self.assertEqual(len(booze), 1)
            self.assertEqual(len(booze[0]), 1)
            self.assertEqual(booze[0][0], 'Victoria Bitter')
        finally:
            con.close()

    def test_description(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            self.assertEqual(cur.description, None,
                             'cursor.description should be none after executing a '
                             'statement that can return no rows (such as DDL)'
                             )
            cur.execute('select name from %sbooze' % self.table_prefix)
            self.assertEqual(len(cur.description), 1,
                             'cursor.description describes too many columns'
                             )
            self.assertEqual(len(cur.description[0]), 7,
                             'cursor.description[x] tuples must have 7 elements'
                             )
            self.assertEqual(cur.description[0][0].lower(), 'name',
                             'cursor.description[x][0] must return column name'
                             )
            self.assertEqual(cur.description[0][1], pymonetdb.STRING,
                             'cursor.description[x][1] must return column type. Got %r'
                             % cur.description[0][1]
                             )

            # Make sure self.description gets reset
            self.executeDDL2(cur)
            self.assertEqual(cur.description, None,
                             'cursor.description not being set to None when executing '
                             'no-result statements (eg. DDL)'
                             )
        finally:
            con.close()

    def test_rowcount(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            self.assertEqual(cur.rowcount, -1,
                             'cursor.rowcount should be -1 after executing no-result '
                             'statements'
                             )
            cur.execute("insert into %sbooze values ('Victoria Bitter')" % (
                self.table_prefix
            ))
            self.assertTrue(cur.rowcount in (-1, 1),
                            'cursor.rowcount should == number or rows inserted, or '
                            'set to -1 after executing an insert statement'
                            )
            cur.execute("select name from %sbooze" % self.table_prefix)
            self.assertTrue(cur.rowcount in (-1, 1),
                            'cursor.rowcount should == number of rows returned, or '
                            'set to -1 after executing a select statement'
                            )
            self.executeDDL2(cur)
            self.assertEqual(cur.rowcount, -1,
                             'cursor.rowcount not being reset to -1 after executing '
                             'no-result statements'
                             )
        finally:
            con.close()

    lower_func = 'lower'

    def test_callproc(self):
        con = self._connect()
        try:
            cur = con.cursor()
            if self.lower_func and hasattr(cur, 'callproc'):
                r = cur.callproc(self.lower_func, ('FOO',))
                self.assertEqual(len(r), 1)
                self.assertEqual(r[0], 'FOO')
                r = cur.fetchall()
                self.assertEqual(len(r), 1, 'callproc produced no result set')
                self.assertEqual(len(r[0]), 1,
                                 'callproc produced invalid result set'
                                 )
                self.assertEqual(r[0][0], 'foo',
                                 'callproc produced invalid results'
                                 )
        finally:
            con.close()

    def test_close(self):
        con = self._connect()
        try:
            cur = con.cursor()
        finally:
            con.close()

        # cursor.execute should raise an Error if called after connection
        # closed
        self.assertRaises(pymonetdb.Error, self.executeDDL1, cur)

        # connection.commit should raise an Error if called after connection'
        # closed.'
        self.assertRaises(pymonetdb.Error, con.commit)

        # connection.close should raise an Error if called more than once
        self.assertRaises(pymonetdb.Error, con.close)

    def test_execute(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self._paraminsert(cur)
        finally:
            con.close()

    def _paraminsert(self, cur):
        self.executeDDL1(cur)
        cur.execute("insert into %sbooze values ('Victoria Bitter')" % (
            self.table_prefix
        ))
        self.assertTrue(cur.rowcount in (-1, 1))

        if pymonetdb.paramstyle == 'qmark':
            cur.execute(
                'insert into %sbooze values (?)' % self.table_prefix,
                ("Cooper's",)
            )
        elif pymonetdb.paramstyle == 'numeric':
            cur.execute(
                'insert into %sbooze values (:1)' % self.table_prefix,
                ("Cooper's",)
            )
        elif pymonetdb.paramstyle == 'named':
            cur.execute(
                'insert into %sbooze values (:beer)' % self.table_prefix,
                {'beer': "Cooper's"}
            )
        elif pymonetdb.paramstyle == 'format':
            cur.execute(
                'insert into %sbooze values (%%s)' % self.table_prefix,
                ("Cooper's",)
            )
        elif pymonetdb.paramstyle == 'pyformat':
            cur.execute(
                'insert into %sbooze values (%%(beer)s)' % self.table_prefix,
                {'beer': "Cooper's"}
            )
        else:
            self.fail('Invalid paramstyle')
        self.assertTrue(cur.rowcount in (-1, 1))

        cur.execute('select name from %sbooze' % self.table_prefix)
        res = cur.fetchall()
        self.assertEqual(len(res), 2, 'cursor.fetchall returned too few rows')
        beers = [res[0][0], res[1][0]]
        beers.sort()
        self.assertEqual(beers[0], "Cooper's",
                         'cursor.fetchall retrieved incorrect data, or data inserted '
                         'incorrectly'
                         )
        self.assertEqual(beers[1], "Victoria Bitter",
                         'cursor.fetchall retrieved incorrect data, or data inserted '
                         'incorrectly'
                         )

    def test_executemany(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            largs = [("Cooper's",), ("Boag's",)]
            margs = [{'beer': "Cooper's"}, {'beer': "Boag's"}]
            if pymonetdb.paramstyle == 'qmark':
                cur.executemany(
                    'insert into %sbooze values (?)' % self.table_prefix,
                    largs
                )
            elif pymonetdb.paramstyle == 'numeric':
                cur.executemany(
                    'insert into %sbooze values (:1)' % self.table_prefix,
                    largs
                )
            elif pymonetdb.paramstyle == 'named':
                cur.executemany(
                    'insert into %sbooze values (:beer)' % self.table_prefix,
                    margs
                )
            elif pymonetdb.paramstyle == 'format':
                cur.executemany(
                    'insert into %sbooze values (%%s)' % self.table_prefix,
                    largs
                )
            elif pymonetdb.paramstyle == 'pyformat':
                cur.executemany(
                    'insert into %sbooze values (%%(beer)s)' % (
                        self.table_prefix
                    ),
                    margs
                )
            else:
                self.fail('Unknown paramstyle')
            self.assertTrue(cur.rowcount in (-1, 2),
                            'insert using cursor.executemany set cursor.rowcount to '
                            'incorrect value %r' % cur.rowcount
                            )
            cur.execute('select name from %sbooze' % self.table_prefix)
            res = cur.fetchall()
            self.assertEqual(len(res), 2,
                             'cursor.fetchall retrieved incorrect number of rows'
                             )
            beers = [res[0][0], res[1][0]]
            beers.sort()
            self.assertEqual(beers[0], "Boag's", 'incorrect data retrieved')
            self.assertEqual(beers[1], "Cooper's", 'incorrect data retrieved')
        finally:
            con.close()

    def test_fetchone(self):
        con = self._connect()
        try:
            cur = con.cursor()

            # cursor.fetchone should raise an Error if called before
            # executing a select-type query
            self.assertRaises(pymonetdb.Error, cur.fetchone)

            # cursor.fetchone should raise an Error if called after
            # executing a query that cannot return rows
            self.executeDDL1(cur)
            self.assertRaises(pymonetdb.Error, cur.fetchone)

            cur.execute('select name from %sbooze' % self.table_prefix)
            self.assertEqual(cur.fetchone(), None,
                             'cursor.fetchone should return None if a query retrieves '
                             'no rows'
                             )
            self.assertTrue(cur.rowcount in (-1, 0))

            # cursor.fetchone should raise an Error if called after
            # executing a query that cannot return rows
            cur.execute("insert into %sbooze values ('Victoria Bitter')" % (
                self.table_prefix
            ))
            self.assertRaises(pymonetdb.Error, cur.fetchone)

            cur.execute('select name from %sbooze' % self.table_prefix)
            r = cur.fetchone()
            self.assertEqual(len(r), 1,
                             'cursor.fetchone should have retrieved a single row'
                             )
            self.assertEqual(r[0], 'Victoria Bitter',
                             'cursor.fetchone retrieved incorrect data'
                             )
            self.assertEqual(cur.fetchone(), None,
                             'cursor.fetchone should return None if no more rows available'
                             )
            self.assertTrue(cur.rowcount in (-1, 1))
        finally:
            con.close()

    samples = [
        'Carlton Cold',
        'Carlton Draft',
        'Mountain Goat',
        'Redback',
        'Victoria Bitter',
        'XXXX'
    ]

    def _populate(self):
        """ Return a list of sql commands to setup the DB for the fetch
            tests.
        """
        populate = [
            "insert into %sbooze values ('%s')" % (self.table_prefix, s)
            for s in self.samples
        ]
        return populate

    def test_cursor_next(self):
        con = self._connect()
        try:
            cur = con.cursor()

            self.executeDDL1(cur)
            for sql in self._populate():
                cur.execute(sql)

            cur.execute('select name from %sbooze' % self.table_prefix)

            for _ in cur:
                pass

        except TypeError:
            self.fail("Cursor iterator not implemented correctly")

        finally:
            con.close()

    def test_fetchmany(self):
        con = self._connect()
        try:
            cur = con.cursor()

            # cursor.fetchmany should raise an Error if called without
            # issuing a query
            self.assertRaises(pymonetdb.Error, cur.fetchmany, 4)

            self.executeDDL1(cur)
            for sql in self._populate():
                cur.execute(sql)

            cur.execute('select name from %sbooze' % self.table_prefix)
            # our default is not one since that is slow
            cur.arraysize = 1
            r = cur.fetchmany()
            self.assertEqual(len(r), 1,
                             'cursor.fetchmany retrieved incorrect number of rows, '
                             'default of arraysize is one.'
                             )
            cur.arraysize = 10
            r = cur.fetchmany(3)  # Should get 3 rows
            self.assertEqual(len(r), 3,
                             'cursor.fetchmany retrieved incorrect number of rows'
                             )
            r = cur.fetchmany(4)  # Should get 2 more
            self.assertEqual(len(r), 2,
                             'cursor.fetchmany retrieved incorrect number of rows'
                             )
            r = cur.fetchmany(4)  # Should be an empty sequence
            self.assertEqual(len(r), 0,
                             'cursor.fetchmany should return an empty sequence after '
                             'results are exhausted'
                             )
            self.assertTrue(cur.rowcount in (-1, 6))

            # Same as above, using cursor.arraysize
            cur.arraysize = 4
            cur.execute('select name from %sbooze' % self.table_prefix)
            r = cur.fetchmany()  # Should get 4 rows
            self.assertEqual(len(r), 4,
                             'cursor.arraysize not being honoured by fetchmany'
                             )
            r = cur.fetchmany()  # Should get 2 more
            self.assertEqual(len(r), 2)
            r = cur.fetchmany()  # Should be an empty sequence
            self.assertEqual(len(r), 0)
            self.assertTrue(cur.rowcount in (-1, 6))

            cur.arraysize = 6
            cur.execute('select name from %sbooze' % self.table_prefix)
            rows = cur.fetchmany()  # Should get all rows
            self.assertTrue(cur.rowcount in (-1, 6))
            self.assertEqual(len(rows), 6)
            self.assertEqual(len(rows), 6)
            rows = [r[0] for r in rows]
            rows.sort()

            # Make sure we get the right data back out
            for i in range(0, 6):
                self.assertEqual(rows[i], self.samples[i],
                                 'incorrect data retrieved by cursor.fetchmany'
                                 )

            rows = cur.fetchmany()  # Should return an empty list
            self.assertEqual(len(rows), 0,
                             'cursor.fetchmany should return an empty sequence if '
                             'called after the whole result set has been fetched'
                             )
            self.assertTrue(cur.rowcount in (-1, 6))

            self.executeDDL2(cur)
            cur.execute('select name from %sbarflys' % self.table_prefix)
            r = cur.fetchmany()  # Should get empty sequence
            self.assertEqual(len(r), 0,
                             'cursor.fetchmany should return an empty sequence if '
                             'query retrieved no rows'
                             )
            self.assertTrue(cur.rowcount in (-1, 0))

        finally:
            con.close()

    def test_fetchall(self):
        con = self._connect()
        try:
            cur = con.cursor()
            # cursor.fetchall should raise an Error if called
            # without executing a query that may return rows (such
            # as a select)
            self.assertRaises(pymonetdb.Error, cur.fetchall)

            self.executeDDL1(cur)
            for sql in self._populate():
                cur.execute(sql)

            # cursor.fetchall should raise an Error if called
            # after executing a a statement that cannot return rows
            self.assertRaises(pymonetdb.Error, cur.fetchall)

            cur.execute('select name from %sbooze' % self.table_prefix)
            rows = cur.fetchall()
            self.assertTrue(cur.rowcount in (-1, len(self.samples)))
            self.assertEqual(len(rows), len(self.samples),
                             'cursor.fetchall did not retrieve all rows'
                             )
            rows = [r[0] for r in rows]
            rows.sort()
            for i in range(0, len(self.samples)):
                self.assertEqual(rows[i], self.samples[i],
                                 'cursor.fetchall retrieved incorrect rows'
                                 )
            rows = cur.fetchall()
            self.assertEqual(
                len(rows), 0,
                'cursor.fetchall should return an empty list if called '
                'after the whole result set has been fetched'
            )
            self.assertTrue(cur.rowcount in (-1, len(self.samples)))

            self.executeDDL2(cur)
            cur.execute('select name from %sbarflys' % self.table_prefix)
            rows = cur.fetchall()
            self.assertTrue(cur.rowcount in (-1, 0))
            self.assertEqual(len(rows), 0,
                             'cursor.fetchall should return an empty list if '
                             'a select query returns no rows'
                             )

        finally:
            con.close()

    def test_mixedfetch(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            for sql in self._populate():
                cur.execute(sql)

            cur.execute('select name from %sbooze' % self.table_prefix)
            rows1 = cur.fetchone()
            rows23 = cur.fetchmany(2)
            rows4 = cur.fetchone()
            rows56 = cur.fetchall()
            self.assertTrue(cur.rowcount in (-1, 6))
            self.assertEqual(len(rows23), 2,
                             'fetchmany returned incorrect number of rows'
                             )
            self.assertEqual(len(rows56), 2,
                             'fetchall returned incorrect number of rows'
                             )

            rows = [rows1[0]]
            rows.extend([rows23[0][0], rows23[1][0]])
            rows.append(rows4[0])
            rows.extend([rows56[0][0], rows56[1][0]])
            rows.sort()
            for i in range(0, len(self.samples)):
                self.assertEqual(rows[i], self.samples[i],
                                 'incorrect data retrieved or inserted'
                                 )
        finally:
            con.close()

    def test_arraysize(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.assertTrue(hasattr(cur, 'arraysize'),
                            'cursor.arraysize must be defined'
                            )
        finally:
            con.close()

    def test_setinputsizes(self):
        con = self._connect()
        try:
            cur = con.cursor()
            cur.setinputsizes((25,))
            self._paraminsert(cur)  # Make sure cursor still works
        finally:
            con.close()

    def test_setoutputsize_basic(self):
        # Basic test is to make sure setoutputsize doesn't blow up
        con = self._connect()
        try:
            cur = con.cursor()
            cur.setoutputsize(1000)
            cur.setoutputsize(2000, 0)
            self._paraminsert(cur)  # Make sure the cursor still works
        finally:
            con.close()

    def test_None(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            cur.execute('insert into %sbooze values (NULL)' % self.table_prefix)
            cur.execute('select name from %sbooze' % self.table_prefix)
            r = cur.fetchall()
            self.assertEqual(len(r), 1)
            self.assertEqual(len(r[0]), 1)
            self.assertEqual(r[0][0], None, 'NULL value not returned as None')
        finally:
            con.close()

    def test_Date(self):
        d1 = pymonetdb.Date(2002, 12, 25)
        d2 = pymonetdb.DateFromTicks(time.mktime((2002, 12, 25, 0, 0, 0, 0, 0, 0)))
        self.assertEqual(str(d1), str(d2))

    def test_Time(self):
        t1 = pymonetdb.Time(13, 45, 30)
        t2 = pymonetdb.TimeFromTicks(time.mktime((2001, 1, 1, 13, 45, 30, 0, 0, 0)))
        self.assertEqual(str(t1), str(t2))

    def test_Timestamp(self):
        t1 = pymonetdb.Timestamp(2002, 12, 25, 13, 45, 30)
        t2 = pymonetdb.TimestampFromTicks(
            time.mktime((2002, 12, 25, 13, 45, 30, 0, 0, 0))
        )
        self.assertEqual(str(t1), str(t2))

    def test_Binary(self):
        b1 = b'\x00\x01\x02\x03'
        b2 = pymonetdb.Binary(b1)
        self.assertEqual(b1, b2)

    def test_STRING(self):
        self.assertTrue(hasattr(pymonetdb, 'STRING'), 'module.STRING must be defined')

    def test_BINARY(self):
        self.assertTrue(hasattr(pymonetdb, 'BINARY'), 'module.BINARY must be defined.')

    def test_NUMBER(self):
        self.assertTrue(hasattr(pymonetdb, 'NUMBER'), 'module.NUMBER must be defined.')

    def test_DATETIME(self):
        self.assertTrue(hasattr(pymonetdb, 'DATETIME'), 'module.DATETIME must be defined.')

    def test_ROWID(self):
        self.assertTrue(hasattr(pymonetdb, 'ROWID'), 'module.ROWID must be defined.')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
import unittest
from unittest.mock import patch
from typing import Optional
from pymonetdb.profiler import ProfilerConnection
from tests import util


class ProfilerTest(unittest.TestCase):
    """Test the profiler connection."""

    conn: Optional[ProfilerConnection] = None

    @classmethod
    def setUpClass(cls):
        cls.conn = ProfilerConnection()
        cls.conn.connect(**util.test_args)

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.conn:
            cls.conn.close()

    @patch('pymonetdb.mapi.Connection._getblock')
    def test_profiler_connection(self, mock_getblock):
        response = '{"key":"value"}\n'  # A random JSON object
        mock_getblock.return_value = response
        self.assertEqual(self.conn.read_object(), response[:-1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
"""
This implements a connection to the profiler.
"""

from pymonetdb import mapi
from pymonetdb.exceptions import OperationalError


class ProfilerConnection:
    """
    A connection to the MonetDB profiler.
    """

    def __init__(self):
        self._mapi = mapi.Connection()
        self._heartbeat = 0
        self._buffer = ""
        self._objects = list()

    def connect(self, database, username="monetdb", password="monetdb", hostname=None, port=50000, heartbeat=0):
        self._heartbeat = heartbeat
        self._mapi.connect(database, username, password, "mal", hostname, port)
        self._mapi.cmd("profiler.setheartbeat(%d);\n" % heartbeat)
        try:
            self._mapi.cmd("profiler.openstream();\n")
        except OperationalError:
            # We might be talking to an older version of MonetDB. Try connecting
            # the old way.
            self._mapi.cmd("profiler.openstream(3);\n")

    def read_object(self):
        self._buffer = self._mapi._getblock()
        while not self._buffer.endswith("}\n"):
            self._buffer += self._mapi._getblock()

        return self._buffer[:-1]

    def close(self):
        if self._mapi:
            self._mapi.disconnect()
            self._mapi = None



import unittest
from pymonetdb import Connection
from tests.util import test_args


class TestNamedTuple(unittest.TestCase):
    def test_namedtuple(self):
        con = Connection(autocommit=False, **test_args)

        cur = con.cursor()

        cur.execute('select 1.0 as floatje')
        cur.fetchone()
        descr = cur.description[0]
        self.assertTrue(descr.name in 'floatje')
        self.assertEqual(descr.type_code, 'decimal')
        self.assertEqual(descr.display_size, None)
        self.assertEqual(type(descr.internal_size), int)
        self.assertEqual(type(descr.precision), int)
        self.assertEqual(type(descr.scale), int)
        self.assertEqual(descr.null_ok, None)

        cur.execute('select 1 as intje')
        cur.fetchone()
        descr = cur.description[0]
        self.assertTrue(descr.name in 'intje')
        self.assertEqual(descr.type_code, 'tinyint')
        self.assertEqual(descr.display_size, None)
        self.assertEqual(type(descr.internal_size), int)
        self.assertEqual(descr.precision, None)
        self.assertEqual(descr.scale, None)
        self.assertEqual(descr.null_ok, None)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import platform
import logging
from pymonetdb import mapi
from pymonetdb.exceptions import OperationalError, InterfaceError


logger = logging.getLogger(__name__)


def parse_statusline(line):
    """
    parses a sabdb format status line. Support v1 and v2.

    """
    if line.startswith("="):
        line = line[1:]
    if not line.startswith('sabdb:'):
        raise OperationalError('wrong result received')

    code, prot_version, rest = line.split(":", 2)

    if prot_version not in ["1", "2"]:
        raise InterfaceError("unsupported sabdb protocol")
    else:
        prot_version = int(prot_version)

    subparts = rest.split(',')
    sub_iter = iter(subparts)

    info = {}

    info['name'] = next(sub_iter)
    info['path'] = next(sub_iter)
    info['locked'] = next(sub_iter) == "1"
    info['state'] = int(next(sub_iter))
    info['scenarios'] = next(sub_iter).split("'")
    if prot_version == 1:
        next(sub_iter)
    info['start_counter'] = int(next(sub_iter))
    info['stop_counter'] = int(next(sub_iter))
    info['crash_counter'] = int(next(sub_iter))
    info['avg_uptime'] = int(next(sub_iter))
    info['max_uptime'] = int(next(sub_iter))
    info['min_uptime'] = int(next(sub_iter))
    info['last_crash'] = int(next(sub_iter))
    info['last_start'] = int(next(sub_iter))
    if prot_version > 1:
        info['last_stop'] = int(next(sub_iter))
    info['crash_avg1'] = next(sub_iter) == "1"
    info['crash_avg10'] = float(next(sub_iter))
    info['crash_avg30'] = float(next(sub_iter))

    return info


def isempty(result):
    """ raises an exception if the result is not empty"""
    if result != "":
        raise OperationalError(result)
    else:
        return True


class Control:
    """
    Use this module to manage your MonetDB databases. You can create, start,
    stop, lock, unlock, destroy your databases and request status information.
    """
    def __init__(self, hostname=None, port=50000, passphrase=None,
                 unix_socket=None, connect_timeout=-1):

        if not unix_socket:
            unix_socket = "/tmp/.s.merovingian.%i" % port

        if platform.system() == "Windows" and not hostname:
            hostname = "localhost"

        self.server = mapi.Connection()
        self.hostname = hostname
        self.port = port
        self.passphrase = passphrase
        self.unix_socket = unix_socket
        self.connect_timeout = connect_timeout

        # check connection
        self.server.connect(hostname=hostname, port=port, username='monetdb',
                            password=passphrase,
                            database='merovingian', language='control',
                            unix_socket=unix_socket,
                            connect_timeout=connect_timeout)
        self.server.disconnect()

    def _send_command(self, database_name, command):
        logger.info("sending '{}' command to database {}".format(command, database_name))
        self.server.connect(hostname=self.hostname, port=self.port,
                            username='monetdb', password=self.passphrase,
                            database='merovingian', language='control',
                            unix_socket=self.unix_socket,
                            connect_timeout=self.connect_timeout)
        result = self.server.cmd("%s %s\n" % (database_name, command))
        self.server.disconnect()
        return result

    def create(self, database_name):
        """
        Initialises a new database or multiplexfunnel in the MonetDB Server.
        A database created with this command makes it available  for use,
        however in maintenance mode (see pymonetdb lock).
        """
        return isempty(self._send_command(database_name, "create"))

    def destroy(self, database_name):
        """
        Removes the given database, including all its data and
        logfiles.  Once destroy has completed, all data is lost.
        Be careful when using this command.
        """
        return isempty(self._send_command(database_name, "destroy"))

    def lock(self, database_name):
        """
        Puts the given database in maintenance mode.  A database
        under maintenance can only be connected to by the DBA.
        A database which is under maintenance is not started
        automatically.  Use the "release" command to bring
        the database back for normal usage.
        """
        return isempty(self._send_command(database_name, "lock"))

    def release(self, database_name):
        """
        Brings back a database from maintenance mode.  A released
        database is available again for normal use.  Use the
        "lock" command to take a database under maintenance.
        """
        return isempty(self._send_command(database_name, "release"))

    def status(self, database_name=False):
        """
        Shows the state of a given glob-style database match, or
        all known if none given.  Instead of the normal mode, a
        long and crash mode control what information is displayed.
        """
        if database_name:
            raw = self._send_command(database_name, "status")
            return parse_statusline(raw)
        else:
            raw = self._send_command("#all", "status")
            return [parse_statusline(line) for line in raw.split("\n")]

    def start(self, database_name):
        """
        Starts the given database, if the MonetDB Database Server
        is running.
        """
        return isempty(self._send_command(database_name, "start"))

    def stop(self, database_name):
        """
        Stops the given database, if the MonetDB Database Server
        is running.
        """
        return isempty(self._send_command(database_name, "stop"))

    def kill(self, database_name):
        """
        Kills the given database, if the MonetDB Database Server
        is running.  Note: killing a database should only be done
        as last resort to stop a database.  A database being
        killed may end up with data loss.
        """
        return isempty(self._send_command(database_name, "kill"))

    def set(self, database_name, property_, value):
        """
        sets property to value for the given database
        for a list of properties, use `pymonetdb get all`
        """
        return isempty(self._send_command(database_name, "%s=%s" % (property_,
                                                                    value)))

    def get(self, database_name):
        """
        gets value for property for the given database, or
        retrieves all properties for the given database
        """
        properties = self._send_command(database_name, "get")
        values = {}
        for line in properties.split("\n"):
            if line.startswith("="):
                line = line[1:]
            if not line.startswith("#"):
                if "=" in line:
                    split = line.split("=")
                    values[split[0]] = split[1]
        return values

    def inherit(self, database_name, property_):
        """
        unsets property, reverting to its inherited value from
        the default configuration for the given database
        """
        return isempty(self._send_command(database_name, property_ + "="))

    def rename(self, old, new):
        return self.set(old, "name", new)

    def defaults(self):
        return self.get("#defaults")

    def neighbours(self):
        return self._send_command("anelosimus", "eximius")



from unittest import TestCase
from pymonetdb import connect
from tests.util import test_args


class TestOid(TestCase):
    def setUp(self):
        self.connection = connect(autocommit=False, **test_args)
        self.cursor = self.connection.cursor()

    def test_oid(self):
        q = "select tag from sys.queue()"
        self.cursor.execute(q)
        self.cursor.fetchall()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from datetime import datetime, timedelta, timezone
import unittest
import pymonetdb.sql.pythonize
import pymonetdb
from tests.util import test_args


class TestPythonize(unittest.TestCase):
    TEST_TIMEZONE = -4

    def setUp(self):
        db = pymonetdb.connect(autocommit=False, **test_args)
        db.set_timezone(self.TEST_TIMEZONE * 3600)
        self.connection = db
        self.cursor = db.cursor()

    def tearDown(self):
        self.connection.close()

    def test_Binary(self):
        input1 = bytes(range(256)).hex()
        output1 = bytes(range(256))
        result1 = pymonetdb.sql.pythonize.convert(input1, pymonetdb.types.BLOB)
        self.assertEqual(output1, result1)

        input2 = b'\tdharma'.hex()
        output2 = b'\tdharma'

        result2 = pymonetdb.sql.pythonize.convert(input2, pymonetdb.types.BLOB)
        self.assertEqual(output2, result2)

    def test_month_interval(self):
        self.cursor.execute('CREATE TEMPORARY TABLE foo (i INTERVAL MONTH)')
        self.cursor.execute('INSERT INTO foo VALUES (INTERVAL \'2\' YEAR)')
        self.cursor.execute('SELECT * from FOO')
        row = self.cursor.fetchone()
        self.assertEqual(row[0], 24)

    def test_timestamptz(self):
        tz = timezone(timedelta(hours=self.TEST_TIMEZONE))
        now = datetime.now(tz)
        self.cursor.execute('SELECT now()')
        row = self.cursor.fetchone()
        ts = row[0]

        # ts is timezone-aware, see the datetime docs
        self.assertIsNotNone(ts.tzinfo)
        self.assertIsNotNone(ts.tzinfo.utcoffset(ts))

        # ts is correct, allowing for fairly large clock skew between client and server
        self.assertAlmostEqual(60 * ts.hour + ts.minute, 60 * now.hour + now.minute, delta=12)

    def test_roundtrip_datetime(self):
        dt = datetime(2020, 2, 14, 20, 50)
        tz = timezone(timedelta(hours=self.TEST_TIMEZONE))
        dtz = dt.replace(tzinfo=tz)

        self.cursor.execute('SELECT %s, %s', [dt, dtz])
        row = self.cursor.fetchone()

        self.assertEqual(row[0].isoformat(), dt.isoformat())
        self.assertEqual(row[1].isoformat(), dtz.isoformat())

    def test_roundtrip_binary(self):
        raw = b'BLUB\x00BLOB'
        wrapped = pymonetdb.Binary(raw)
        self.cursor.execute('SELECT %s, %s', [raw, wrapped])
        row = self.cursor.fetchone()

        self.assertEqual(row[0], raw)
        self.assertEqual(row[1], raw)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
import typing

from pymonetdb.exceptions import ProgrammingError
from .uploads import Uploader, Upload
from .downloads import Downloader, Download
from .directoryhandler import SafeDirectoryHandler

if typing.TYPE_CHECKING:
    from pymonetdb.mapi import Connection

# these are used in the code but they are referred to in the docs
(Uploader, Downloader, SafeDirectoryHandler)


def handle_file_transfer(mapi: "Connection", cmd: str):
    if cmd.startswith("r "):
        parts = cmd[2:].split(' ', 2)
        if len(parts) == 2:
            try:
                n = int(parts[0])
            except ValueError:
                pass
            return handle_upload(mapi, parts[1], True, n)
    elif cmd.startswith("rb "):
        return handle_upload(mapi, cmd[2:], False, 0)
    elif cmd.startswith("w "):
        return handle_download(mapi, cmd[2:], True)
    elif cmd.startswith("wb "):
        return handle_download(mapi, cmd[3:], False)
    else:
        pass
    # we only reach this if decoding the cmd went wrong:
    mapi._putblock(f"Invalid file transfer command: {cmd!r}")


def handle_upload(mapi: "Connection", filename: str, text_mode: bool, offset: int):
    if not mapi.uploader:
        mapi._putblock("No upload handler has been registered with pymonetdb\n")
        return
    skip_amount = offset - 1 if offset > 0 else 0
    upload = Upload(mapi)
    try:
        mapi.uploader.handle_upload(upload, filename, text_mode, skip_amount)
    except Exception as e:
        # We must make sure the server doesn't think this is a succesful upload.
        # The protocol does not allow us to flag an error after the upload has started,
        # so the only thing we can do is kill the connection
        upload.error = True
        mapi._sabotage()
        raise e
    finally:
        upload.close()
    if not upload.has_been_used():
        raise ProgrammingError("Upload handler didn't do anything")


def handle_download(mapi: "Connection", filename: str, text_mode: bool):
    if not mapi.downloader:
        mapi._putblock("No download handler has been registered with pymonetdb\n")
        return
    download = Download(mapi)
    try:
        mapi.downloader.handle_download(download, filename, text_mode)
    except Exception as e:
        # For consistency we also drop the connection on these exceptions.
        #
        # # Alternatively we might just discard the incoming data and allow
        # work to continue, but in 99% of the cases the application is about
        # to crash and it makes no sense to delay that by first reading all
        # the data.
        #
        # Also, if the download has not really started yet we might send
        # an error message to the server but then you get inconsistent
        # behaviour: if the download hadn't started yet, the transaction ends
        # up in an aborted state and must be ROLLed BACK, but if the download
        # has started we discard all data and allow it to continue without
        # error.
        #
        # Bottom line is that it's easier to understand if we just always
        # crash the connection.
        download._shutdown()
        mapi._sabotage()
        raise e
    finally:
        download.close()



"""
This is the python implementation of the mapi protocol.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import socket
import logging
import struct
import hashlib
import os
import typing
from typing import Optional, Tuple
from urllib.parse import urlparse, parse_qs

from pymonetdb.exceptions import OperationalError, DatabaseError, \
    ProgrammingError, NotSupportedError, IntegrityError

if typing.TYPE_CHECKING:
    from pymonetdb.filetransfer.downloads import Downloader
    from pymonetdb.filetransfer.uploads import Uploader

logger = logging.getLogger(__name__)

MAX_PACKAGE_LENGTH = (1024 * 8) - 2

MSG_PROMPT = ""
MSG_MORE = "\1\2\n"
MSG_FILETRANS = "\1\3\n"
MSG_INFO = "#"
MSG_ERROR = "!"
MSG_Q = "&"
MSG_QTABLE = "&1"
MSG_QUPDATE = "&2"
MSG_QSCHEMA = "&3"
MSG_QTRANS = "&4"
MSG_QPREPARE = "&5"
MSG_QBLOCK = "&6"
MSG_HEADER = "%"
MSG_TUPLE = "["
MSG_TUPLE_NOSLICE = "="
MSG_REDIRECT = "^"
MSG_OK = "=OK"

MSG_FILETRANS_B = bytes(MSG_FILETRANS, 'utf-8')

STATE_INIT = 0
STATE_READY = 1

# MonetDB error codes
errors = {
    '42S02': OperationalError,  # no such table
    '40002': IntegrityError,  # INSERT INTO: UNIQUE constraint violated
    '2D000': IntegrityError,  # COMMIT: failed
    '40000': IntegrityError,  # DROP TABLE: FOREIGN KEY constraint violated
    'M0M29': IntegrityError,  # The code monetdb emitted before Jun2020
}


def handle_error(error):
    """Return exception matching error code.

    args:
        error (str): error string, potentially containing mapi error code

    returns:
        tuple (Exception, formatted error): returns OperationalError if unknown
            error or no error code in string

    """

    if error[:13] == 'SQLException:':
        idx = str.index(error, ':', 14)
        error = error[idx + 10:]
    if len(error) > 5 and error[:5] in errors:
        return errors[error[:5]], error
    else:
        return OperationalError, error


# noinspection PyExceptionInherit
class Connection(object):
    """
    MAPI (low level MonetDB API) connection
    """

    def __init__(self):
        self.state = STATE_INIT
        self._result = None
        self.socket: Optional[socket.socket] = None
        self.unix_socket = None
        self.hostname = ""
        self.port = 0
        self.username = ""
        self.database = ""
        self.language = ""
        self.handshake_options = None
        self.connect_timeout = socket.getdefaulttimeout()
        self.uploader = None
        self.downloader = None
        self.stashed_buffer = None

    def connect(self, database: str, username: str, password: str, language: str,  # noqa: C901
                hostname: Optional[str] = None, port: Optional[int] = None, unix_socket=None, connect_timeout=-1,
                handshake_options=None):
        """ setup connection to MAPI server

        unix_socket is used if hostname is not defined.
        """

        if ':' in database:
            if not database.startswith('mapi:monetdb:'):
                raise DatabaseError("colon not allowed in database name, except as part of "
                                    "mapi:monetdb://<hostname>[:<port>]/<database> URI")
            parsed = urlparse(database[5:])
            # parse basic settings
            if parsed.hostname or parsed.port:
                # connect over tcp
                if not parsed.path.startswith('/'):
                    raise DatabaseError('invalid mapi url')
                database = parsed.path[1:]
                if '/' in database:
                    raise DatabaseError('invalid mapi url')
                username = parsed.username or username
                password = parsed.password or password
                hostname = parsed.hostname or hostname
                port = parsed.port or port
            else:
                # connect over unix domain socket
                unix_socket = parsed.path or unix_socket
                username = parsed.username or username
                password = parsed.password or password
                database = ''  # must be set in uri parameter
            # parse uri parameters
            if parsed.query:
                parms = parse_qs(parsed.query)
                if 'database' in parms:
                    if database == '':
                        database = parms['database'][-1]
                    else:
                        raise DatabaseError('database= query parameter is only allowed with unix domain sockets')
                # Future work: parse other parameters such as reply_size.

        if hostname and hostname[:1] == '/' and not unix_socket:
            unix_socket = f'{hostname}/.s.monetdb.{port}'
            hostname = None
        if not unix_socket and os.path.exists(f"/tmp/.s.monetdb.{port}"):
            unix_socket = f"/tmp/.s.monetdb.{port}"
        elif not unix_socket and not hostname:
            hostname = 'localhost'

        # None and zero are allowed values
        if connect_timeout != -1:
            assert connect_timeout is None or connect_timeout >= 0
            self.connect_timeout = connect_timeout

        self.hostname = hostname
        self.port = port
        self.username = username
        self.database = database
        self.language = language
        self.unix_socket = unix_socket
        self.handshake_options = handshake_options or []
        if hostname:
            if self.socket:
                self.socket.close()
                self.socket = None
            for af, socktype, proto, canonname, sa in socket.getaddrinfo(hostname, port,
                                                                         socket.AF_UNSPEC, socket.SOCK_STREAM):
                try:
                    self.socket = socket.socket(af, socktype, proto)
                    # For performance, mirror MonetDB/src/common/stream.c socket settings.
                    self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                    self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                    self.socket.settimeout(self.connect_timeout)
                except socket.error as msg:
                    logger.debug(f"'{msg}' for af {af} with socktype {socktype}")
                    self.socket = None
                    continue
                try:
                    self.socket.connect(sa)
                except socket.error as msg:
                    logger.info(msg.strerror)
                    self.socket.close()
                    self.socket = None
                    continue
                break
            if self.socket is None:
                raise socket.error("Connection refused")
        else:
            self.socket = socket.socket(socket.AF_UNIX)
            self.socket.settimeout(self.connect_timeout)
            self.socket.connect(unix_socket)
            if self.language != 'control':
                # don't know why, but we need to do this
                self.socket.send('0'.encode())

        if not (self.language == 'control' and not self.hostname):
            # control doesn't require authentication over socket
            self._login(password=password)

        self.socket.settimeout(socket.getdefaulttimeout())
        self.state = STATE_READY

    def _login(self, password: str, iteration=0):
        """ Reads challenge from line, generate response and check if
        everything is okay """

        challenge = self._getblock()
        response = self._challenge_response(challenge, password)
        self._putblock(response)
        prompt = self._getblock().strip()

        if len(prompt) == 0:
            # Empty response, server is happy
            pass
        elif prompt == MSG_OK:
            pass
        elif prompt.startswith(MSG_INFO):
            logger.info("%s" % prompt[1:])

        elif prompt.startswith(MSG_ERROR):
            logger.error(prompt[1:])
            raise DatabaseError(prompt[1:])

        elif prompt.startswith(MSG_REDIRECT):
            # a redirect can contain multiple redirects, for now we only use
            # the first
            redirect = prompt.split()[0][1:].split(':')
            if redirect[1] == "merovingian":
                logger.debug("restarting authentication")
                if iteration <= 10:
                    self._login(iteration=iteration + 1, password=password)
                else:
                    raise OperationalError("maximal number of redirects "
                                           "reached (10)")

            elif redirect[1] == "monetdb":
                self.hostname = redirect[2][2:]
                self.port, self.database = redirect[3].split('/')
                self.port = int(self.port)
                logger.info("redirect to monetdb://%s:%s/%s" %
                            (self.hostname, self.port, self.database))
                if self.socket:
                    self.socket.close()
                self.connect(hostname=self.hostname, port=self.port,
                             username=self.username, password=password,
                             database=self.database, language=self.language)

            else:
                raise ProgrammingError("unknown redirect: %s" % prompt)

        else:
            raise ProgrammingError("unknown state: %s" % prompt)

    def disconnect(self):
        """ disconnect from the monetdb server """
        logger.info("disconnecting from database")
        self.state = STATE_INIT
        self.socket.close()

    def _sabotage(self):
        """ Kill the connection in a way that the server is sure to recognize as an error"""
        sock = self.socket
        self.socket = None
        self.state = STATE_INIT
        if not sock:
            return
        bad_header = struct.pack('<H', 2 * 8193 + 0)  # larger than allowed, and not the final message
        bad_body = b"ERROR\x80ERROR"  # invalid utf-8, and too small
        try:
            sock.send(bad_header + bad_body)
            # and then we hang up
            sock.close()
        except Exception:
            # don't care
            pass

    def cmd(self, operation):  # noqa: C901
        """ put a mapi command on the line"""
        logger.debug("executing command %s" % operation)

        if self.state != STATE_READY:
            raise ProgrammingError("Not connected")

        self._putblock(operation)
        response = self._getblock_and_transfer_files()
        if not len(response):
            return ""
        elif response.startswith(MSG_OK):
            return response[3:].strip() or ""
        if response == MSG_MORE:
            # tell server it isn't going to get more
            return self.cmd("")

        # If we are performing an update test for errors such as a failed
        # transaction.

        # We are splitting the response into lines and checking each one if it
        # starts with MSG_ERROR. If this is the case, find which line records
        # the error and use it to call handle_error.
        if response[:2] == MSG_QUPDATE:
            lines = response.split('\n')
            if any([line.startswith(MSG_ERROR) for line in lines]):
                index = next(i for i, v in enumerate(lines) if v.startswith(MSG_ERROR))
                exception, msg = handle_error(lines[index][1:])
                raise exception(msg)

        if response[0] in [MSG_Q, MSG_HEADER, MSG_TUPLE]:
            return response
        elif response[0] == MSG_ERROR:
            exception, msg = handle_error(response[1:])
            raise exception(msg)
        elif response[0] == MSG_INFO:
            logger.info("%s" % (response[1:]))
        elif self.language == 'control' and not self.hostname:
            if response.startswith("OK"):
                return response[2:].strip() or ""
            else:
                return response
        else:
            raise ProgrammingError("unknown state: %s" % response)

    def _challenge_response(self, challenge: str, password: str):  # noqa: C901
        """ generate a response to a mapi login challenge """

        challenges = challenge.split(':')
        if challenges[-1] != '' or len(challenges) < 7:
            raise OperationalError("Server sent invalid challenge")
        challenges.pop()

        salt, identity, protocol, hashes, endian = challenges[:5]

        if protocol == '9':
            algo = challenges[5]
            try:
                h = hashlib.new(algo)
                h.update(password.encode())
                password = h.hexdigest()
            except ValueError as e:
                raise NotSupportedError(str(e))
        else:
            raise NotSupportedError("We only speak protocol v9")

        for i in hashes.split(","):
            try:
                s = hashlib.new(i)
            except ValueError:
                pass
            else:
                s.update(password.encode())
                s.update(salt.encode())
                pwhash = "{" + i + "}" + s.hexdigest()
                break
        else:
            raise NotSupportedError("Unsupported hash algorithms required"
                                    " for login: %s" % hashes)

        response = ":".join(["BIG", self.username, pwhash, self.language, self.database]) + ":"

        if len(challenges) >= 7:
            response += "FILETRANS:"
            options_level = 0
            for part in challenges[6].split(","):
                if part.startswith("sql="):
                    try:
                        options_level = int(part[4:])
                    except ValueError:
                        raise OperationalError("invalid sql options level in server challenge: " + part)
            options = []
            for opt in self.handshake_options:
                if opt.level < options_level:
                    options.append(opt.name + "=" + str(int(opt.value)))
                    opt.sent = True
            response += ",".join(options) + ":"

        return response

    def _getblock_and_transfer_files(self):
        """ read one mapi encoded block and take care of any file transfers the server requests"""
        buffer = self._get_buffer()
        offset = 0

        # import this here to solve circular import
        from pymonetdb.filetransfer import handle_file_transfer

        while True:
            old_offset = offset
            offset = self._getblock_raw(buffer, old_offset)
            i = buffer.rfind(b'\n', old_offset, offset - 1)
            if i >= old_offset + 2 and buffer[i - 2: i + 1] == MSG_FILETRANS_B:
                # File transfer request. Chop the cmd off the buffer by lowering the offset
                cmd = str(buffer[i + 1: offset - 1], 'utf-8')
                offset = i - 2
                handle_file_transfer(self, cmd)
                continue
            else:
                break
        self._stash_buffer(buffer)
        return str(memoryview(buffer)[:offset], 'utf-8')

    def _getblock(self) -> str:
        """ read one mapi encoded block """
        buf = self._get_buffer()
        end = self._getblock_raw(buf, 0)
        ret = str(memoryview(buf)[:end], 'utf-8')
        self._stash_buffer(buf)
        return ret

    def _getblock_raw(self, buffer: bytearray, offset: int) -> int:
        """
        Read one mapi block into 'buffer' starting at 'offset', enlarging the buffer
        as necessary and returning offset plus the number of bytes read.
        """
        last = False
        while not last:
            offset, last = self._get_minor_block(buffer, offset)
        return offset

    def _get_minor_block(self, buffer: bytearray, offset: int) -> Tuple[int, bool]:
        self._getbytes(buffer, offset, 2)
        unpacked = buffer[offset] + 256 * buffer[offset + 1]
        length = unpacked >> 1
        last = unpacked & 1
        if length:
            offset = self._getbytes(buffer, offset, length)
        return (offset, bool(last))

    def _getbytes(self, buffer: bytearray, offset: int, count: int) -> int:
        """
        Read 'count' bytes from the socket into 'buffer' starting at 'offset'.
        Enlarge buffer if necessary.
        Return offset + count if all goes well.
        """
        assert self.socket
        end = count + offset
        if len(buffer) < end:
            # enlarge
            nblocks = 1 + (end - len(buffer)) // 8192
            buffer += bytes(nblocks * 8192)
        while offset < end:
            view = memoryview(buffer)[offset:end]
            n = self.socket.recv_into(view)
            if n == 0:
                raise BrokenPipeError("Server closed connection")
            offset += n
        return end

    def _get_buffer(self) -> bytearray:
        """Retrieve a previously stashed buffer for reuse, or create a new one"""
        if self.stashed_buffer:
            buffer = self.stashed_buffer
            self.stashed_buffer = None
        else:
            buffer = bytearray(8192)
        return buffer

    def _stash_buffer(self, buffer):
        """Stash a used buffer for future reuse"""
        if self.stashed_buffer is None or len(self.stashed_buffer) < len(buffer):
            self.stashed_buffer = buffer

    def _putblock(self, block):
        """ wrap the line in mapi format and put it into the socket """
        self._putblock_inet_raw(block.encode(), True)

    def _putblock_raw(self, block, finish: bool):
        """ put the data into the socket """
        self._putblock_inet_raw(block, finish)

    def _putblock_inet_raw(self, block, finish):
        pos = 0
        last = 0
        while not last:
            data = memoryview(block)[pos:pos + MAX_PACKAGE_LENGTH]
            length = len(data)
            if length < MAX_PACKAGE_LENGTH:
                last = 1
            flag = struct.pack('<H', (length << 1) + (last if finish else 0))
            self.socket.send(flag)
            self.socket.send(data)
            pos += length

    def __del__(self):
        if self.socket:
            self.socket.close()

    def set_reply_size(self, size):
        # type: (int) -> None
        """
        Set the amount of rows returned by the server.

        args:
            size: The number of rows
        """

        self.cmd("Xreply_size %s" % size)

    def set_uploader(self, uploader: "Uploader"):
        """Register the given Uploader, or None to deregister"""
        self.uploader = uploader

    def set_downloader(self, downloader: "Downloader"):
        """Register the given Downloader, or None to deregister"""
        self.downloader = downloader


# When all supported Python versions support it we can enable @dataclass here.
class HandshakeOption:
    """
    Option that can be set during the MAPI handshake

    Should be sent as <name>=<val>, where <val> is `value` converted to int.
    The `level` is used to determine if the server supports this option.
    The `fallback` is a function-like object that can be called with the
    value (not converted to an integer) as a parameter.
    Field `sent` can be used to keep track of whether the option has been sent.
    """
    def __init__(self, level, name, fallback, value):
        self.level = level
        self.name = name
        self.value = value
        self.fallback = fallback
        self.sent = False



#!/usr/bin/env python3

import os
import platform
import subprocess
import sys
import time

import pytest


def start_mserver(monetdbdir, farmdir, dbname, port, logfile):  # noqa: C901
    exe = os.path.join(monetdbdir, 'bin', 'mserver5')
    if platform.system() == 'Windows':
        exe += '.exe'
    dbpath = os.path.join(farmdir, dbname)
    try:
        os.mkdir(dbpath)
    except FileExistsError:
        pass
    #
    env = dict((k, v) for k, v in os.environ.items())
    path_components = [
        os.path.join(monetdbdir, "bin"),
        os.path.join(monetdbdir, "lib", "monetdb5"),
        env['PATH'],
    ]
    env['PATH'] = os.pathsep.join(path_components)
    sets = dict(
        prefix=monetdbdir,
        exec_prefix=monetdbdir,
        mapi_port=port,
    )
    cmdline = [
        exe,
        f'--dbpath={dbpath}',
    ]
    for k, v in sets.items():
        cmdline.append('--set')
        cmdline.append(f'{k}={v}')
    print()
    print('-- Starting mserver')
    print(f'-- PATH={env["PATH"]}')
    print(f'-- cmdline: {cmdline!r}')
    t0 = time.time()
    awkward_silence = t0 + 2
    proc = subprocess.Popen(cmdline, env=env, stderr=open(logfile, 'wb'))
    #
    while True:
        try:
            code = proc.wait(timeout=0.1)
            exit(f'mserver unexpectedly exited with code {code}')
        except subprocess.TimeoutExpired:
            if os.path.exists(os.path.join(dbpath, '.started')):
                break
            t = time.time()
            if t >= awkward_silence:
                print(f"-- Waited for {t - t0:.1f}s")
                awkward_silence = t + 1
            if t > t0 + 30.1:
                print("Starting mserver took too long, giving up")
                proc.kill()
                exit("given up")
    print('-- mserver has started')
    return proc


if len(sys.argv) != 5:
    exit(f"Usage: {sys.argv[0]} MONETDIR FARMDIR DBNAME PORT")
monet_dir = sys.argv[1]
farm_dir = sys.argv[2]
db_name = sys.argv[3]
db_port = int(sys.argv[4])

proc = start_mserver(monet_dir, farm_dir, db_name, db_port, os.path.join(farm_dir, "errlog"))
try:
    print('The reported default encoding is', sys.getdefaultencoding())
    with open(os.path.join(farm_dir, 'w.txt'), 'w') as f:
        print("The encoding for 'w' files is", f.encoding)
    with open(os.path.join(farm_dir, 'w.txt'), 'wt') as f:
        print("The encoding for 'wt files is", f.encoding)
    ret = pytest.main(args=['-k', 'not test_control'])
    exit(ret)
finally:
    if proc.returncode is None:
        print('-- Killing the server')
        proc.kill()
    else:
        print('-- Server has already terminated')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from datetime import datetime, timedelta, timezone
import logging
import platform

from pymonetdb.sql import cursors
from pymonetdb import exceptions
from pymonetdb import mapi

logger = logging.getLogger("pymonetdb")


class Connection:
    """A MonetDB SQL database connection"""
    default_cursor = cursors.Cursor

    def __init__(self, database, hostname=None, port=50000, username="monetdb",
                 password="monetdb", unix_socket=None, autocommit=False,
                 host=None, user=None, connect_timeout=-1):
        """ Set up a connection to a MonetDB SQL database.

        args:
            database (str): name of the database, or MAPI URI
            hostname (str): Hostname where monetDB is running
            port (int): port to connect to (default: 50000)
            username (str): username for connection (default: "monetdb")
            password (str): password for connection (default: "monetdb")
            unix_socket (str): socket to connect to. used when hostname not set
                                (default: "/tmp/.s.monetdb.50000")
            autocommit (bool):  enable/disable auto commit (default: False)
            connect_timeout -- the socket timeout while connecting
                               (default: see python socket module)

        MAPI URI:
            tcp socket:         mapi:monetdb://[<username>[:<password>]@]<host>[:<port>]/<database>
            unix domain socket: mapi:monetdb:///[<username>[:<password>]@]path/to/socket?database=<database>

        returns:
            Connection object

        """

        self.autocommit = autocommit
        self.sizeheader = True
        self.replysize = 0

        # The DB API spec is not specific about this
        if host:
            hostname = host
        if user:
            username = user

        if platform.system() == "Windows" and not hostname:
            hostname = "localhost"

        # Level numbers taken from mapi.h.
        # The options start out with member .sent set to False.
        handshake_options = [
            mapi.HandshakeOption(1, "auto_commit", self.set_autocommit, autocommit),
            mapi.HandshakeOption(2, "reply_size", self.set_replysize, 100),
            mapi.HandshakeOption(3, "size_header", self.set_sizeheader, True),
            mapi.HandshakeOption(5, "time_zone", self.set_timezone, _local_timezone_offset_seconds()),
        ]

        self.mapi = mapi.Connection()
        self.mapi.connect(hostname=hostname, port=int(port), username=username,
                          password=password, database=database, language="sql",
                          unix_socket=unix_socket, connect_timeout=connect_timeout,
                          handshake_options=handshake_options)

        # self.mapi.connect() has set .sent to True for all items that
        # have already been arranged during the initial challenge/response.
        # Now take care of the rest.
        for option in handshake_options:
            if not option.sent:
                option.fallback(option.value)

    def close(self):
        """ Close the connection.

        The connection will be unusable from this
        point forward; an Error exception will be raised if any operation
        is attempted with the connection. The same applies to all cursor
        objects trying to use the connection.  Note that closing a connection
        without committing the changes first will cause an implicit rollback
        to be performed.
        """
        if self.mapi:
            if not self.autocommit:
                self.rollback()
            self.mapi.disconnect()
            self.mapi = None
        else:
            raise exceptions.Error("already closed")

    def set_autocommit(self, autocommit):
        """
        Set auto commit on or off. 'autocommit' must be a boolean
        """
        self.command("Xauto_commit %s" % int(autocommit))
        self.autocommit = autocommit

    def set_sizeheader(self, sizeheader):
        """
        Set sizeheader on or off. When enabled monetdb will return
        the size a type. 'sizeheader' must be a boolean.
        """
        self.command("Xsizeheader %s" % int(sizeheader))
        self.sizeheader = sizeheader

    def set_replysize(self, replysize):
        self.command("Xreply_size %s" % int(replysize))
        self.replysize = replysize

    def set_timezone(self, seconds_east_of_utc):
        hours = int(seconds_east_of_utc / 3600)
        remaining = seconds_east_of_utc - 3600 * hours
        minutes = int(remaining / 60)
        cmd = f"SET TIME ZONE INTERVAL '{hours:+03}:{abs(minutes):02}' HOUR TO MINUTE;"
        c = self.cursor()
        c.execute(cmd)
        c.close()

    def set_uploader(self, uploader):
        """
        Register an Uploader object which will handle file upload requests.

        Must be an instance of class pymonetdb.Uploader or None.
        """
        self.mapi.set_uploader(uploader)

    def set_downloader(self, downloader):
        """
        Register a Downloader object which will handle file download requests.

        Must be an instance of class pymonetdb.Downloader or None
        """
        self.mapi.set_downloader(downloader)

    def commit(self):
        """
        Commit any pending transaction to the database. Note that
        if the database supports an auto-commit feature, this must
        be initially off. An interface method may be provided to
        turn it back on.

        Database modules that do not support transactions should
        implement this method with void functionality.
        """
        self.__mapi_check()
        return self.cursor().execute('COMMIT')

    def rollback(self):
        """
        This method is optional since not all databases provide
        transaction support.

        In case a database does provide transactions this method
        causes the database to roll back to the start of any
        pending transaction.  Closing a connection without
        committing the changes first will cause an implicit
        rollback to be performed.
        """
        self.__mapi_check()
        return self.cursor().execute('ROLLBACK')

    def cursor(self):
        """
        Return a new Cursor Object using the connection.  If the
        database does not provide a direct cursor concept, the
        module will have to emulate cursors using other means to
        the extent needed by this specification.
        """
        return cursors.Cursor(self)

    def execute(self, query):
        """ use this for executing SQL queries """
        return self.command('s' + query + '\n;')

    def command(self, command):
        """ use this function to send low level mapi commands """
        self.__mapi_check()
        return self.mapi.cmd(command)

    def __mapi_check(self):
        """ check if there is a connection with a server """
        if not self.mapi:
            raise exceptions.Error("connection closed")
        return True

    def settimeout(self, timeout):
        """ set the amount of time before a connection times out """
        self.mapi.socket.settimeout(timeout)

    def gettimeout(self):
        """ get the amount of time before a connection times out """
        return self.mapi.socket.gettimeout()

    # these are required by the python DBAPI
    Warning = exceptions.Warning
    Error = exceptions.Error
    InterfaceError = exceptions.InterfaceError
    DatabaseError = exceptions.DatabaseError
    DataError = exceptions.DataError
    OperationalError = exceptions.OperationalError
    IntegrityError = exceptions.IntegrityError
    InternalError = exceptions.InternalError
    ProgrammingError = exceptions.ProgrammingError
    NotSupportedError = exceptions.NotSupportedError


def _local_timezone_offset_seconds():
    # local time
    our_now = datetime.now().replace(microsecond=0).astimezone()
    # same year/month/day/hour/min/etc, but marked as UTC
    utc_now = our_now.replace(tzinfo=timezone(timedelta(0)))
    # UTC reaches a given hour/min/seconds combination later than
    # the time zones east of UTC do. This means the offset is
    # positive if we are east.
    return round(utc_now.timestamp() - our_now.timestamp())



"""
Classes related to file transfer requests as used by COPY INTO ON CLIENT.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
import typing
from io import BufferedIOBase, BufferedWriter, RawIOBase, TextIOBase, TextIOWrapper
from abc import ABC, abstractmethod
from typing import Any, Optional, Union
from pymonetdb.mapi import MSG_MORE, MSG_FILETRANS
from pymonetdb.exceptions import ProgrammingError

if typing.TYPE_CHECKING:
    from pymonetdb.mapi import Connection as MapiConnection


class Upload:
    """
    Represents a request from the server to upload data to the server. It is
    passed to the Uploader registered by the application, which for example
    might retrieve the data from a file on the client system. See
    pymonetdb.sql.connections.Connection.set_uploader().

    Use the method send_error() to refuse the upload, binary_writer() to get a
    binary file object to write to, or text_writer() to get a text-mode file
    object to write to.

    Implementations should be VERY CAREFUL to validate the file name before
    opening any files on the client system!
    """

    mapi: Optional["MapiConnection"]
    error = False
    cancelled = False
    bytes_sent = 0
    chunk_size = 1024 * 1024
    chunk_used = 0
    rawio: Optional["UploadIO"] = None
    writer: Optional[BufferedWriter] = None
    twriter: Optional[TextIOBase] = None

    def __init__(self, mapi: "MapiConnection"):
        self.mapi = mapi

    def _check_usable(self):
        if self.error:
            raise ProgrammingError("Upload handle has had an error, cannot be used anymore")
        if not self.mapi:
            raise ProgrammingError("Upload handle has been closed, cannot be used anymore")

    def is_cancelled(self) -> bool:
        """Returns true if the server has cancelled the upload."""
        return self.cancelled

    def has_been_used(self) -> bool:
        """Returns true if .send_error(), .text_writer() or .binary_writer() have been called."""
        return self.error or (self.rawio is not None)

    def set_chunk_size(self, size: int):
        """
        After every CHUNK_SIZE bytes, the server gets the opportunity to cancel
        the rest of the upload. Defaults to 1 MiB.
        """
        self.chunk_size = size

    def send_error(self, message: str) -> None:
        """
        Tell the server the requested upload has been refused
        """
        if self.cancelled:
            return
        self._check_usable()
        if self.bytes_sent:
            raise ProgrammingError("Cannot send error after data has been sent")
        if not message.endswith("\n"):
            message += "\n"
        self.error = True
        assert self.mapi
        self.mapi._putblock(message)
        self.mapi = None

    def _raw(self) -> "UploadIO":
        if self.bytes_sent == 0:
            # send the magic newline indicating we're ok with the upload
            self._send(b'\n', False)
        if not self.rawio:
            self.rawio = UploadIO(self)
        return self.rawio

    def binary_writer(self) -> BufferedIOBase:
        """
        Returns a binary file-like object. All data written to it is uploaded
        to the server.
        """
        if not self.writer:
            self.writer = BufferedWriter(self._raw())
        return self.writer

    def text_writer(self) -> TextIOBase:
        r"""
        Returns a text-mode file-like object. All text written to it is uploaded
        to the server. DOS/Windows style line endings (CR LF, \\r \\n) are
        automatically rewritten to single \\n's.
        """
        if not self.twriter:
            # Without the Any annotation there is no way I can convince the
            # type checker that TextIOWrapper can accept a NormalizeCrLf
            # object. Apparently being a subclass of BufferedIOBase is not enough.
            w: Any = NormalizeCrLf(self._raw())
            self.twriter = TextIOWrapper(w, encoding='utf-8', newline='\n')
        return self.twriter

    def _send_data(self, data: Union[bytes, memoryview]):
        if self.cancelled:
            return
        self._check_usable()
        assert self.mapi is not None
        pos = 0
        end = len(data)
        while pos < end:
            n = min(end - pos, self.chunk_size - self.chunk_used)
            chunk = memoryview(data)[pos:pos + n]
            if n == self.chunk_size - self.chunk_used and self.chunk_size > 0:
                server_wants_more = self._send_and_get_prompt(chunk)
                if not server_wants_more:
                    self.cancelled = True
                    self.mapi.uploader.cancel()
                    self.mapi = None
                    break
            else:
                self._send(chunk, False)
            pos += n

    def _send(self, data: Union[bytes, memoryview], finish: bool):
        assert self.mapi
        self.mapi._putblock_raw(data, finish)
        self.bytes_sent += len(data)
        self.chunk_used += len(data)

    def _send_and_get_prompt(self, data: Union[bytes, memoryview]) -> bool:
        assert self.mapi
        self._send(data, True)
        prompt = self.mapi._getblock()
        if prompt == MSG_MORE:
            self.chunk_used = 0
            return True
        elif prompt == MSG_FILETRANS:
            # server says stop
            return False
        else:
            raise ProgrammingError(f"Unexpected server response: {prompt[:50]!r}")

    def close(self):
        """
        End the upload succesfully
        """
        if self.error:
            return
        if self.twriter:
            self.twriter.close()
        if self.writer:
            self.writer.close()
        if self.mapi:
            server_wants_more = False
            if self.chunk_used != 0:
                # finish the current block
                server_wants_more = self._send_and_get_prompt(b'')
            if server_wants_more:
                # send empty block to indicate end of upload
                self.mapi._putblock('')
                # receive acknowledgement
                resp = self.mapi._getblock()
                if resp != MSG_FILETRANS:
                    raise ProgrammingError(f"Unexpected server response: {resp[:50]!r}")
            self.mapi = None


class UploadIO(RawIOBase):
    """IO adaptor for Upload. """

    def __init__(self, upload: Upload):
        self.upload = upload

    def writable(self):
        return True

    def write(self, b):
        n = len(b)
        if self.upload.is_cancelled():
            return n
        self.upload._send_data(b)
        return n


class NormalizeCrLf(BufferedIOBase):
    """
    Helper class used to normalize line endings before sending text to MonetDB.

    Existing normalization code mostly deals with normalizing after reading,
    this one normalizes before writing.
    """

    def __init__(self, inner):
        self.inner = inner
        self.pending = False

    def writable(self):
        return True

    def write(self, data) -> int:
        if not data:
            return 0

        if self.pending:
            if data.startswith(b"\n"):
                # normalize by forgetting the pending \r
                pass
            else:
                # pending \r not followed by \n, write it
                self.inner.write(b"\r")
                # do not take the above write into account in the return value,
                # it was included last time

        normalized = data.replace(b"\r\n", b"\n")

        if normalized[-1] == 13:  # \r
            # not sure if it will be followed by \n, move it to pending
            self.pending = True
            normalized = memoryview(normalized)[:-1]
        else:
            self.pending = False

        n = self.inner.write(normalized)
        assert n == len(normalized)
        return len(data)

    def flush(self):
        return self.inner.flush()

    def close(self):
        if self.pending:
            self.inner.write(b"\r")
            self.pending = False
        return self.inner.close()


class Uploader(ABC):
    """
    Base class for upload hooks. Instances of subclasses of this class can be
    registered using pymonetdb.Connection.set_uploader(). Every time an upload
    request is received, an Upload object is created and passed to this objects
    .handle_upload() method.

    If the server cancels the upload halfway, the .cancel() methods is called
    and all further data written is ignored.
    """

    @abstractmethod
    def handle_upload(self, upload: Upload, filename: str, text_mode: bool,
                      skip_amount: int):
        """
        Called when an upload request is received. Implementations should either
        send an error using upload.send_error(), or request a writer using
        upload.text_writer() or upload.binary_writer(). All data written to the
        writer will be sent to the server.

        Parameter 'filename' is the file name used in the COPY INTO statement.
        Parameter 'text_mode' indicates whether the server requested a text file
        or a binary file. In case of a text file, 'skip_amount' indicates the
        number of lines to skip. In binary mode, 'skip_amount' is always 0.

        SECURITY NOTE! Make sure to carefully validate the file name before
        opening files on the file system. Otherwise, if an adversary has taken
        control of the network connection or of the server, they can use file
        upload requests to read arbitrary files from your computer
        (../../)
        """
        pass

    def cancel(self):
        """Optional method called when the server cancels the upload."""
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

"""
functions for converting monetdb SQL fields to Python objects
"""

import json
import time
import datetime
import re
import uuid
from decimal import Decimal
from datetime import timedelta, timezone

from pymonetdb.sql import types
from pymonetdb.exceptions import ProgrammingError


def _extract_timezone(data):
    sign_symbol = data[-6]

    if sign_symbol == '+':
        sign = 1
    elif sign_symbol == '-':
        sign = -1
    else:
        raise ProgrammingError("no + or - in %s" % data)

    hours = sign * int(data[-5:-3])
    minutes = sign * int(data[-2:])
    delta = timedelta(hours=hours, minutes=minutes)
    timezone = datetime.timezone(delta)

    return data[:-6], timezone


def strip(data):
    """ returns a python string, with chopped off quotes,
    and replaced escape characters"""
    return ''.join([w.encode('utf-8').decode('unicode_escape')
                    if '\\' in w else w
                    for w in re.split('([\000-\200]+)', data[1:-1])])


def py_bool(data):
    """ return python boolean """
    return data == "true"


def py_time(data):
    """ returns a python Time
    """
    if '.' in data:
        return datetime.datetime.strptime(data, '%H:%M:%S.%f').time()
    else:
        return datetime.datetime.strptime(data, '%H:%M:%S').time()


def py_timetz(data):
    """ returns a python Time where data contains a tz code
    """
    t, timezone_delta = _extract_timezone(data)
    if '.' in t:
        return datetime.datetime.strptime(t, '%H:%M:%S.%f').time().replace(tzinfo=timezone_delta)
    else:
        return datetime.datetime.strptime(t, '%H:%M:%S').time().replace(tzinfo=timezone_delta)


def py_date(data):
    """ Returns a python Date
    """
    return datetime.datetime.strptime(data, '%Y-%m-%d').date()


def py_timestamp(data):
    """ Returns a python Timestamp
    """
    if '.' in data:
        return datetime.datetime.strptime(data, '%Y-%m-%d %H:%M:%S.%f')
    else:
        return datetime.datetime.strptime(data, '%Y-%m-%d %H:%M:%S')


def py_timestamptz(data):
    """ Returns a python Timestamp where data contains a tz code
    """
    dt, timezone_delta = _extract_timezone(data)
    if '.' in dt:
        return datetime.datetime.strptime(dt, '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=timezone_delta)
    else:
        return datetime.datetime.strptime(dt, '%Y-%m-%d %H:%M:%S').replace(tzinfo=timezone_delta)


def py_sec_interval(data: str) -> timedelta:
    """ Returns a python TimeDelta where data represents a value of MonetDB's INTERVAL SECOND type
    which resembles a stringified decimal.
    """
    return timedelta(seconds=int(Decimal(data)))


def py_day_interval(data: str) -> int:
    """ Returns a python number of days where data represents a value of MonetDB's INTERVAL DAY type
    which resembles a stringified decimal.
    """
    return timedelta(seconds=int(Decimal(data))).days


def py_bytes(data: str):
    """Returns a bytes (py3) or string (py2) object representing the input blob."""
    return bytes.fromhex(data)


def oid(data):
    """represents an object identifier

    For now we will just return the string representation just like mclient does.
    """
    return oid


mapping = {
    types.CHAR: strip,
    types.VARCHAR: strip,
    types.CLOB: strip,
    types.BLOB: py_bytes,
    types.TINYINT: int,
    types.SMALLINT: int,
    types.INT: int,
    types.BIGINT: int,
    types.HUGEINT: int,
    types.SERIAL: int,
    types.SHORTINT: int,
    types.MEDIUMINT: int,
    types.LONGINT: int,
    types.OID: oid,
    types.WRD: int,
    types.REAL: float,
    types.FLOAT: float,
    types.DOUBLE: float,
    types.DECIMAL: Decimal,
    types.BOOLEAN: py_bool,
    types.DATE: py_date,
    types.TIME: py_time,
    types.TIMESTAMP: py_timestamp,
    types.TIMETZ: py_timetz,
    types.TIMESTAMPTZ: py_timestamptz,
    types.MONTH_INTERVAL: int,
    types.SEC_INTERVAL: py_sec_interval,
    types.DAY_INTERVAL: py_day_interval,
    types.URL: strip,
    types.INET: str,
    types.UUID: uuid.UUID,
    types.JSON: json.loads,
    types.GEOMETRY: strip,
    types.GEOMETRYA: strip,
    types.MBR: strip,
    types.XML: str,
}


def convert(data, type_code):
    """
    Calls the appropriate convertion function based upon the python type
    """

    # null values should always be replaced by None type
    if data == "NULL":
        return None
    try:
        return mapping[type_code](data)
    except KeyError:
        raise ProgrammingError("type %s is not supported" % type_code)


# below stuff required by the DBAPI

def Binary(data):
    """Convert to wraps binary data"""
    assert isinstance(data, bytes) or isinstance(data, bytearray)
    return data


def DateFromTicks(ticks):
    """Convert ticks to python Date"""
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    """Convert ticks to python Time"""
    return Time(*time.localtime(ticks)[3:6])


def TimeTzFromTicks(ticks):
    """Convert ticks to python Time"""
    return _make_localtime(Time(*time.localtime(ticks)[3:6]))


def TimestampFromTicks(ticks):
    """Convert ticks to python Timestamp"""
    return Timestamp(*time.localtime(ticks)[:6])


def TimestampTzFromTicks(ticks):
    """Convert ticks to python Timestamp"""
    return _make_localtime(Timestamp(*time.localtime(ticks)[:6]))


def _make_localtime(t):
    return t.replace(tzinfo=timezone(timedelta(hours=1)))


Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime
STRING = types.VARCHAR
BINARY = types.BLOB
NUMBER = types.DECIMAL
DATE = types.DATE
TIME = types.TIME
DATETIME = types.TIMESTAMP
ROWID = types.INT



"""
Classes related to file transfer requests as used by COPY INTO ON CLIENT.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


import codecs
from importlib import import_module
from pathlib import Path
from shutil import copyfileobj
from typing import Optional
from pymonetdb.filetransfer.uploads import Upload, Uploader
from pymonetdb.filetransfer.downloads import Download, Downloader


class SafeDirectoryHandler(Uploader, Downloader):
    """
    File transfer handler which uploads and downloads files from a given
    directory, taking care not to allow access to files outside that directory.
    Instances of this class can be registered using the pymonetb.Connection's
    set_uploader() and set_downloader() methods.

    When downloading text files, the downloaded text is converted according to
    the `encoding` and `newline` parameters, if present. Valid values for
    `encoding` are any encoding known to Python, or None. Valid values for
    `newline` are `"\\\\n"`, `"\\\\r\\\\n"` or None. None means to use the
    system default.

    For binary up- and downloads, no conversions are applied.

    When uploading text files, the `encoding` parameter indicates how the text
    is read and `newline` is mostly ignored: both `\\\\n` and `\\\\r\\\\n` are
    valid line endings. The exception is that because the server expects its
    input to be `\\\\n`-terminated UTF-8 text,  if you set encoding to "utf-8"
    and newline to "\\\\n", text mode transfers are performed as binary, which
    improves performance. For uploads, only do this if you are absolutely,
    positively sure that all files in the directory are actually valid UTF-8
    encoded and have Unix line endings.

    If `compression` is set to True, which is the default, the
    SafeDirectoryHandler will automatically compress and decompress files with
    extensions .gz, .bz2, .xz and .lz4. Note that the first three algorithms are
    built into Python, but LZ4 only works if the lz4.frame module is available.
    """

    def __init__(self, dir, encoding: Optional[str] = None, newline: Optional[str] = None, compression=True):
        self.dir = Path(dir).resolve()
        self.encoding = encoding
        self.is_utf8 = (self.encoding and (codecs.lookup('utf-8') == codecs.lookup(self.encoding)))
        self.newline = newline
        self.compression = compression

    def secure_resolve(self, filename: str) -> Optional[Path]:
        p = self.dir.joinpath(filename).resolve()
        if str(p).startswith(str(self.dir)):
            return p
        else:
            return None

    def handle_upload(self, upload: Upload, filename: str, text_mode: bool, skip_amount: int):  # noqa: C901
        """:meta private:"""  # keep the API docs cleaner, this has already been documented on class Uploader.

        p = self.secure_resolve(filename)
        if not p:
            return upload.send_error("Forbidden")

        if self.is_utf8 and self.newline == "\n" and skip_amount == 0:
            # optimization
            text_mode = False

        # open
        if text_mode:
            mode = "rt"
            encoding = self.encoding
            newline = self.newline
        else:
            mode = "rb"
            encoding = None
            newline = None
        try:
            opener = lookup_compression_algorithm(filename) if self.compression else open
        except ModuleNotFoundError as e:
            return upload.send_error(str(e))
        try:
            f = opener(p, mode=mode, encoding=encoding, newline=newline)
        except IOError as e:
            return upload.send_error(str(e))

        with f:
            if text_mode:
                tw = upload.text_writer()
                for _ in range(skip_amount):
                    if not f.readline():
                        break
                self._upload_data(upload, f, tw)
            else:
                bw = upload.binary_writer()
                self._upload_data(upload, f, bw)

    def _upload_data(self, upload: Upload, src, dst):
        # Due to duck typing this method works equally well in text- and binary mode
        bufsize = 1024 * 1024
        while not upload.is_cancelled():
            data = src.read(bufsize)
            if not data:
                break
            dst.write(data)

    def handle_download(self, download: Download, filename: str, text_mode: bool):
        p = self.secure_resolve(filename)
        if not p:
            return download.send_error("Forbidden")

        if self.is_utf8 and self.newline == "\n":
            # optimization
            text_mode = False

        # open
        mode = "w" if text_mode else "wb"
        if text_mode:
            mode = "wt"
            encoding = self.encoding
            newline = self.newline
        else:
            mode = "wb"
            encoding = None
            newline = None
        try:
            opener = lookup_compression_algorithm(filename) if self.compression else open
        except ModuleNotFoundError as e:
            return download.send_error(str(e))
        try:
            f = opener(p, mode=mode, encoding=encoding, newline=newline)
        except IOError as e:
            return download.send_error(str(e))

        with f:
            if text_mode:
                tr = download.text_reader()
                copyfileobj(tr, f)
            else:
                br = download.binary_reader()
                copyfileobj(br, f)


def lookup_compression_algorithm(filename: str):
    lowercase = str(filename).lower()
    if lowercase.endswith('.gz'):
        mod = 'gzip'
    elif lowercase.endswith('.bz2'):
        mod = 'bz2'
    elif lowercase.endswith('.xz'):
        mod = 'lzma'
    elif lowercase.endswith('.lz4'):
        # not always available
        mod = 'lz4.frame'
    else:
        return open
    opener = import_module(mod).open
    return opener



"""
This is a MonetDB Python API.

To set up a connection use pymonetdb.connect()

"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import pkg_resources
from pymonetdb import sql
from pymonetdb import mapi
from pymonetdb import exceptions

from pymonetdb.profiler import ProfilerConnection
from pymonetdb.sql.connections import Connection
from pymonetdb.sql.pythonize import BINARY, Binary, DATE, Date, Time, Timestamp, DateFromTicks, TimestampFromTicks, \
    TimeFromTicks, NUMBER, ROWID, STRING, TIME, types, DATETIME, TimeTzFromTicks, TimestampTzFromTicks
from pymonetdb.exceptions import Error, DataError, DatabaseError, IntegrityError, InterfaceError, InternalError, \
    NotSupportedError, OperationalError, ProgrammingError, Warning
from pymonetdb.filetransfer.downloads import Download, Downloader
from pymonetdb.filetransfer.uploads import Upload, Uploader
from pymonetdb.filetransfer.directoryhandler import SafeDirectoryHandler

try:
    __version__ = pkg_resources.require("pymonetdb")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "1.0rc"

apilevel = "2.0"
threadsafety = 1
paramstyle = "pyformat"

__all__ = ['sql', 'mapi', 'exceptions', 'BINARY', 'Binary', 'connect', 'Connection', 'DATE', 'Date', 'Time',
           'Timestamp', 'DateFromTicks', 'TimeFromTicks', 'TimestampFromTicks', 'DataError', 'DatabaseError', 'Error',
           'IntegrityError', 'InterfaceError', 'InternalError', 'NUMBER', 'NotSupportedError', 'OperationalError',
           'ProgrammingError', 'ROWID', 'STRING', 'TIME', 'Warning', 'apilevel', 'connect', 'paramstyle',
           'threadsafety', 'Download', 'Downloader', 'Upload', 'Uploader', 'SafeDirectoryHandler', 'types', 'DATETIME',
           'TimeTzFromTicks', 'TimestampTzFromTicks']


def connect(*args, **kwargs) -> Connection:
    return Connection(*args, **kwargs)


def profiler_connection(*args, **kwargs):
    c = ProfilerConnection()
    c.connect(*args, **kwargs)
    return c


connect.__doc__ = Connection.__init__.__doc__
profiler_connection.__doc__ = ProfilerConnection.__init__.__doc__



"""
Classes related to file transfer requests as used by COPY INTO ON CLIENT.
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.


from abc import ABC, abstractmethod
from io import BufferedIOBase, TextIOWrapper
from pymonetdb import mapi as mapi_protocol
from pymonetdb.exceptions import ProgrammingError


class Download:
    """
    Represents a request from the server to download data from the server. It is
    passed to the Downloader registered by the application, which for example
    might write the data to a file on the client system. See
    pymonetdb.Connection.set_downloader().

    Use the method send_error() to refuse the download, binary_reader() to get a
    binary file object to read bytes from, or text_reader() to get a text-mode
    file object to read text from.

    Implementations should be EXTREMELY CAREFUL to validate the file name before
    opening and writing to any files on the client system!
    """

    def __init__(self, mapi: "mapi_protocol.Connection"):
        self.mapi = mapi
        self.started = False
        self.buffer = bytearray(8190)
        self.pos = 0
        self.len = 0
        self.reader = None
        self.treader = None

    def send_error(self, message: str) -> None:
        """
        Tell the server the requested download is refused
        """
        if self.started:
            raise ProgrammingError("Cannot send error anymore")
        if not self.mapi:
            return
        self.started = True
        if not message.endswith("\n"):
            message += "\n"
        self.mapi._putblock(message)
        self._shutdown()

    def binary_reader(self):
        """Returns a binary file-like object to read the downloaded data from."""
        if not self.reader:
            if not self.mapi:
                raise ProgrammingError("download has already been closed")
            self.started = True
            self.mapi._putblock("\n")
            self.reader = DownloadIO(self)
        return self.reader

    def text_reader(self):
        """Returns a text mode file-like object to read the downloaded data from."""
        if not self.treader:
            self.treader = TextIOWrapper(self.binary_reader(), encoding='utf-8', newline='\n')
        return self.treader

    def close(self):
        """End the download succesfully. Any unconsumed data will be discarded."""
        while self.mapi:
            self._fetch()
        assert not self.mapi

    def _available(self) -> int:
        return self.len - self.pos

    def _consume(self, n: int) -> memoryview:
        end = min(self.pos + n, self.len)
        ret = memoryview(self.buffer)[self.pos:end]
        self.pos = end
        return ret

    def _fetch(self):
        if not self.mapi:
            return
        self.pos = 0
        self.len = 0   # safety in case of exceptions
        self.len, last = self.mapi._get_minor_block(self.buffer, 0)
        if last:
            self._shutdown()

    def _shutdown(self):
        self.started = True
        self.mapi = None


class DownloadIO(BufferedIOBase):

    def __init__(self, download: Download):
        self.download = download

    def readable(self):
        return True

    def read(self, n=0):
        if self.download._available() == 0:
            self.download._fetch()
        return bytes(self.download._consume(n))

    def read1(self, n=0):
        return self.read(n)


class Downloader(ABC):
    """
    Base class for download hooks. Instances of subclasses of this class can be
    registered using pymonetdb.Connection.set_downloader(). Every time a
    download request arrives, a Download object is created and passed to this
    objects .handle_download() method.

    SECURITY NOTE! Make sure to carefully validate the file name before opening
    files on the file system. Otherwise, if an adversary has taken control of
    the network connection or of the server, they can use download requests to
    OVERWRITE ARBITRARY FILES on your computer
    """

    @abstractmethod
    def handle_download(self, download: Download, filename: str, text_mode: bool):
        """
        Called when a download request is received. Implementations should either
        send an error using download.send_error(), or request a reader using
        download.text_reader() or download.binary_reader().

        Parameter 'filename' is the file name used in the COPY INTO statement.
        Parameter 'text_mode' indicates whether the server requested text
        or binary mode.

        SECURITY NOTE! Make sure to carefully validate the file name before
        opening files on the file system. Otherwise, if an adversary has taken
        control of the network connection or of the server, they can use file
        download requests to overwrite arbitrary files on your computer.
        (../../)
        """
        pass



"""
Module containing shared utilities only used for testing MonetDB
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

from importlib import import_module
from os import environ

test_port = int(environ.get('MAPIPORT', 50000))
test_database = environ.get('TSTDB', 'demo')
test_hostname = environ.get('TSTHOSTNAME', 'localhost')
test_username = environ.get('TSTUSERNAME', 'monetdb')
test_password = environ.get('TSTPASSWORD', 'monetdb')
test_passphrase = environ.get('TSTPASSPHRASE', 'testdb')
test_full = environ.get('TSTFULL', 'false').lower() == 'true'

test_args = {
    'port': test_port,
    'database': test_database,
    'hostname': test_hostname,
    'username': test_username,
    'password': test_password,
}


try:
    import_module('lz4.frame')
    have_lz4 = True
except ModuleNotFoundError:
    have_lz4 = False
